/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file enju.cc
 *  Main program of enju
 *
 **********************************************************************/

#include "config_mayz.h"
#include <liblilfes/lilfes.h>
#include <liblilfes/builtin.h>
#include <liblilfes/option.h>
#include <liblilfes/utility.h>
#include <liblilfes/rl_stream.h>
#include "ParserCkyIter.h"
#include "AmisModel.h"

using namespace lilfes;

#ifndef ENJU_DEFAULT_TAGGER
#define ENJU_DEFAULT_TAGGER "uptagger"
#endif // ENJU_DEFAULT_TAGGER

static const std::string ENJU_DATA_ENV( "ENJU_DIR" );
static const std::string ENJU_TAGGER_ENV( "ENJU_TAGGER" );
static const std::string DEFAULT_ENJU_DATA_DIR( ENJU_DIR );
static const std::string LILFES_PATH_ENV( "LILFES_PATH" );
static const std::string DEFAULT_ENJU_TAGGER( ENJU_DEFAULT_TAGGER );
static const std::string DEFAULT_FOM_MODULE( "enju/synmodel" );
static const std::string DEFAULT_GRAMMAR_MODULE( "enju/grammar" );
static const std::string DEFAULT_APPLICATION_MODULE( "enju/outputdep" );
static const std::string DEFAULT_COMMAND( "output_dependency_file." );
static const std::string OUTPUTXML_APPLICATION_MODULE( "enju/outputxml" );
static const std::string OUTPUTXML_COMMAND( "output_xml_file." );
static const std::string OUTPUTSO_COMMAND( "output_so_file." );
static const std::string OUTPUT_SUMMARY_COMMAND( "output_summary_file." );
static const std::string MORIV_APPLICATION_MODULE( "enju/moriv" );
static const std::string MORIV_COMMAND( "moriv." );
static const std::string MORIV_CGI_COMMAND( "cgi." );

static const int BEAM_NUM_START = 12;
static const int BEAM_NUM_STEP = 6;
static const int BEAM_NUM_END = 30;

static const double BEAM_WIDTH_START = 6.0;
static const double BEAM_WIDTH_STEP = 3.0;
static const double BEAM_WIDTH_END = 15.0;

static const double GLOBAL_BEAM_WIDTH_START = 8.0;
static const double GLOBAL_BEAM_WIDTH_STEP = 4.0;
static const double GLOBAL_BEAM_WIDTH_END = 30.0;

static std::string fom_name( DEFAULT_FOM_MODULE );
static std::string grammar_name( DEFAULT_GRAMMAR_MODULE );
static std::string parser_name( "mayz/up" );
static std::string application_name;
static std::string command_name;
static std::vector< std::string > module_list;
static std::vector< std::string > eval_list;
static bool interactive_mode = false;

static size_t limit_sentence_length = 0;
static size_t limit_edge_number = 0;
static int limit_constituent_size = 20;

static bool global_thresholding = true;
static bool iterative_parse = true;

//////////////////////////////////////////////////////////////////////

void startup_message() {
  *error_stream << "Enju " << ENJU_VERSION << "" << std::endl;
  *error_stream << ENJU_COPYRIGHT << std::endl;
}

void help_message() {
  *error_stream << "Usage: enju [options] [-a arguments]" << std::endl;
  *error_stream << "Arguments following \"-a\" are passed to LiLFeS programs as command-line arguments" << std::endl;
  *error_stream << "Options:" << std::endl;
  *error_stream << "  -D dir\tDirectory of enju files (default: " << DEFAULT_ENJU_DATA_DIR << ")" << std::endl;
  *error_stream << "  -L dir\tDirectory of LiLFeS modules" << std::endl;
  *error_stream << "  -t tagger\tPOS tagger (default: " << DEFAULT_ENJU_TAGGER << ")" << std::endl;
  *error_stream << "  -d\t\tOutput in dependency format" << std::endl;
  *error_stream << "  -s\t\tOutput in summary format" << std::endl;
  *error_stream << "  -xml\t\tOutput in xml format" << std::endl;
  *error_stream << "  -so\t\tOutput in xml stand-off format" << std::endl;
  *error_stream << "  -cgi\t\tStart CGI server" << std::endl;
  *error_stream << "  -moriv\tStart MoriV CGI server" << std::endl;
  *error_stream << "  -W number\tLimit number of words" << std::endl;
  *error_stream << "  -E number\tLimit number of edges" << std::endl;
  *error_stream << "  -C number\tLimit size of large constituents" << std::endl;
  *error_stream << "  -l module\tLoad LiLFeS program" << std::endl;
  *error_stream << "  -e command\tExecute LiLFeS command" << std::endl;
  *error_stream << "  -i\t\tInteractive mode (show LiLFeS prompt)" << std::endl;
  *error_stream << "  -n\t\tNon-interactive mode (default)" << std::endl;
  *error_stream << "  -g module\tGrammar module (default: " << DEFAULT_GRAMMAR_MODULE << ")" << std::endl;
  *error_stream << "  -f module\tFOM module (default: " << DEFAULT_FOM_MODULE << ")" << std::endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  for ( int i = 1; i < argc; ++i ) {
    //std::cerr << i << " " << argv[ i ] << std::endl;
    if ( argv[ i ][ 0 ] == '-' ) {
      // "-a" option
      if ( std::strcmp( argv[ i ], "-a" ) == 0 ) {
        lilfes_arg_flag::SetFlagFromStringArray( argc - i - 1, argv + i + 1 );
        break;
      }

      // option without argument
      if ( std::strcmp( argv[ i ], "-d" ) == 0 ) {
        // output in dependency
          application_name = DEFAULT_APPLICATION_MODULE;
          command_name = DEFAULT_COMMAND;
          continue;
      }
      if ( std::strcmp( argv[ i ], "-s" ) == 0 ) {
        // output in summary
          application_name = DEFAULT_APPLICATION_MODULE;
          command_name = OUTPUT_SUMMARY_COMMAND;
          continue;
      }
      if ( std::strcmp( argv[ i ], "-xml" ) == 0 ) {
        // output parse trees and dependencies in xml format
          application_name = OUTPUTXML_APPLICATION_MODULE;
          command_name = OUTPUTXML_COMMAND;
          continue;
      }
      if ( std::strcmp( argv[ i ], "-so" ) == 0 ) {
        // output parse trees and dependencies in xml format
          application_name = OUTPUTXML_APPLICATION_MODULE;
          command_name = OUTPUTSO_COMMAND;
          continue;
      }
      if ( std::strcmp( argv[ i ], "-cgi" ) == 0 ) {
        // moriv server
        application_name = MORIV_APPLICATION_MODULE;
        command_name = MORIV_CGI_COMMAND;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-moriv" ) == 0 ) {
        // moriv server
        application_name = MORIV_APPLICATION_MODULE;
        command_name = MORIV_COMMAND;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-i" ) == 0 ) {
        // interactive mode
        interactive_mode = true;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-n" ) == 0 ) {
        // non-interactive mode
        interactive_mode = false;
        continue;
      }
      if ( std::strcmp( argv[ i ], "--enable-global-thresholding" ) == 0 ) {
        // enable global thresholding
        global_thresholding = true;
        continue;
      }
      if ( std::strcmp( argv[ i ], "--disable-global-thresholding" ) == 0 ) {
        // disable global thresholding
        global_thresholding = false;
        continue;
      }
      if ( std::strcmp( argv[ i ], "--enable-iterative-parse" ) == 0 ) {
        // enable iterative parse
        iterative_parse = true;
        continue;
      }
      if ( std::strcmp( argv[ i ], "--disable-iterative-parse" ) == 0 ) {
        // disable iterative parse
        iterative_parse = false;
        continue;
      }

      // option with argument
      if ( i + 1 < argc ) {
        if ( std::strcmp( argv[ i ], "-D" ) == 0 ) {
          // data path
          setenv( ENJU_DATA_ENV.c_str(), argv[ ++i ], 1 );
          continue;
        }
        if ( std::strcmp( argv[ i ], "-L" ) == 0 ) {
          // data path
          const char* lilfes_path = getenv( LILFES_PATH_ENV.c_str() );
          if ( lilfes_path == NULL || *lilfes_path == '\0' ) {
            setenv( LILFES_PATH_ENV.c_str(), argv[ ++i ], 1 );
          } else {
            std::string lilfes_path_str( argv[ ++i ] );
            lilfes_path_str += ':';
            lilfes_path_str += lilfes_path;
            setenv( LILFES_PATH_ENV.c_str(), lilfes_path_str.c_str(), 1 );
          }
          continue;
        }
        if ( std::strcmp( argv[ i ], "-W" ) == 0 ) {
          // limit sentence length
          char* endptr;
          int len = strtol( argv[ ++i ], &endptr, 0 );
          if ( *endptr != '\0' || len < 0 ) {
            *error_stream << "argument of \"-W\" must be a positive integer" << std::endl;
            return false;
          }
          limit_sentence_length = len;
          continue;
        }
        if ( std::strcmp( argv[ i ], "-E" ) == 0 ) {
          // limit edge number
          char* endptr;
          int num = strtol( argv[ ++i ], &endptr, 0 );
          if ( *endptr != '\0' || num < 0 ) {
            *error_stream << "argument of \"-E\" must be a positive integer" << std::endl;
            return false;
          }
          limit_edge_number = num;
          continue;
        }
        if ( std::strcmp( argv[ i ], "-C" ) == 0 ) {
          // limit size of constituents
          char* endptr;
          int num = strtol( argv[ ++i ], &endptr, 0 );
          if ( *endptr != '\0' || num < 0 ) {
            *error_stream << "argument of \"-C\" must be a positive integer" << std::endl;
            return false;
          }
          limit_constituent_size = num;
          continue;
        }
        if ( std::strcmp( argv[ i ], "-t" ) == 0 ) {
          // external tagger
          setenv( ENJU_TAGGER_ENV.c_str(), argv[ ++i ], 1 );
          continue;
        }
        if ( std::strcmp( argv[ i ], "-l" ) == 0 ) {
          // load module
          module_list.push_back( argv[ ++i ] );
          continue;
        }
        if ( std::strcmp( argv[ i ], "-e" ) == 0 ) {
          // execute command
          eval_list.push_back( argv[ ++i ] );
          continue;
        }
        if ( std::strcmp( argv[ i ], "-g" ) == 0 ) {
          // grammar module
          grammar_name = argv[ ++i ];
          continue;
        }
        if ( std::strcmp( argv[ i ], "-f" ) == 0 ) {
          // FOM module
          fom_name = argv[ ++i ];
          continue;
        }
      }

      *error_stream << "Unknown option: " << argv[ i ] << std::endl;
      return false;

    } else {
      *error_stream << "Unknown argument: " << argv[ i ] << std::endl;
      return false;
    }
  }
  return true;
}

//////////////////////////////////////////////////////////////////////

int main( int argc, char** argv ) {
  try {
    // Initialize everything
    Initializer::PerformAll();

    // Initialize a machine object
    machine mach;

    // show startup message
    startup_message();

    // show help message
    if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
      help_message();
      return 0;
    }
    // set defaults
    setenv( ENJU_DATA_ENV.c_str(), DEFAULT_ENJU_DATA_DIR.c_str(), 0 );
    setenv( ENJU_TAGGER_ENV.c_str(), DEFAULT_ENJU_TAGGER.c_str(), 0 );
    // command-line arguments
    if ( ! analyze_arguments( argc, argv ) ) {
      help_message();
      return 1;
    }
    if ( application_name.empty() && command_name.empty() && module_list.empty() && eval_list.empty() ) {
      // default application module
      application_name = DEFAULT_APPLICATION_MODULE;
      if ( ! interactive_mode ) {
        command_name = DEFAULT_COMMAND;
      }
    }
    if ( ! application_name.empty() ) module_list.push_back( application_name );
    if ( ! command_name.empty() ) eval_list.push_back( command_name );

    // Load grammar and FOM module
    *error_stream << "Loading grammar module \"" << grammar_name << "\"... " << std::flush;
    if ( ! load_module( mach, "grammar module", grammar_name ) ) {
      *error_stream << "not found" << std::endl;
      return 1;
    }
    *error_stream << "done." << std::endl;
    *error_stream << "Loading FOM module \"" << fom_name << "\"... " << std::flush;
    if ( ! load_module( mach, "FOM module", fom_name ) ) {
      *error_stream << "not found" << std::endl;
      return 1;
    }
    *error_stream << "done." << std::endl;
    *error_stream << "Loading parser module \"" << parser_name << "\"... " << std::flush;
    if ( ! load_module( mach, "parser module", parser_name ) ) {
      *error_stream << "not found" << std::endl;
      return 1;
    }
    *error_stream << "done." << std::endl;

    // Load application modules
    for ( std::vector< std::string >::const_iterator it = module_list.begin();
          it != module_list.end();
          ++it ) {
      *error_stream << "Loading module \"" << *it << "\"... " << std::flush;
      if ( ! load_module( mach, "application module", *it ) ) {
        *error_stream << "not found" << std::endl;
        return 1;
      }
      *error_stream << "done." << std::endl;
    }

    // Make a parser
    *error_stream << "Initializing parser..." << std::endl;

    up::Grammar grammar;
    if(! grammar.init( &mach ) ) {
      *error_stream << "initialization of a grammar failed." << std::endl;
      return 1;
    }

    up::ParserCkyIter parser;
    parser.setBeamThresholdNum(BEAM_NUM_START);
    parser.setBeamThresholdWidth(BEAM_WIDTH_START);
    parser.setGlobalBeamThresholdWidth(GLOBAL_BEAM_WIDTH_START);
    parser.enablePreLex();
    if ( iterative_parse ) {
      parser.enableIterativeParse( BEAM_NUM_START, BEAM_WIDTH_START, GLOBAL_BEAM_WIDTH_START,
                                   BEAM_NUM_END, BEAM_WIDTH_END, GLOBAL_BEAM_WIDTH_END,
                                   BEAM_NUM_STEP, BEAM_WIDTH_STEP, GLOBAL_BEAM_WIDTH_STEP);
      parser.enableReuseIterativeParse();
    }
    if ( global_thresholding ) {
      parser.enableGlobalThresholding();
    }
//    parser.enablePriorThresholding();
    if ( limit_constituent_size > 0 ) {
      parser.setLimitConstituentSize( limit_constituent_size );
    }
    if(! parser.init( &mach, &grammar) ) {
        *error_stream << "failed." << std::endl;
        return 1;
    }
    if ( limit_sentence_length > 0 ) {
      parser.setLimitWords( limit_sentence_length );
    }
    if ( limit_edge_number > 0 ) {
      parser.setLimitEdges( limit_edge_number );
    }
    parser.enableQuickCheck();
    
    *error_stream << "done." << std::endl;

    // Start lilfes command
    *error_stream << "Ready" << std::endl;
    module::SetCurrentModule( module::UserModule() );
    for ( std::vector< std::string >::iterator it = eval_list.begin();
          it != eval_list.end();
          ++it ) {
      if ( (*it)[ it->find_last_not_of( " \t" ) ] != '.' ) {
        it->append( "." );
      }
      FSP command = mach.eval( it->c_str() );
      if ( ! command.IsValid() ) {
        *error_stream << "Command not found: " << *it << std::endl;
        return 1;
      }
      //std::cerr << command.DisplayAVM();
      command = command.Follow( hd );  // predicate is in hd:
      if ( ! mach.call( command ) ) {
        *error_stream << "Execution of command failed: " << *it << std::endl;
      }
    }

    // Start interactive mode
    if ( interactive_mode ) {
      *error_stream << "Go into interactive mode" << std::endl;
      lilfes::lexer_t lexer(&lilfes::rl_istream::rl_stdin, "stdin");
      mach.parse(lexer);
    }

    // Termination process
    *error_stream << "Terminating parser..." << std::endl;
    if(! parser.term()) {
        *error_stream << "parser failed to terminate" << std::endl;
        return 1;
    }
    if(! grammar.term()) {
      *error_stream << "termination of the grammar failed" << std::endl;
      return 1;
    }
    *error_stream << "done." << std::endl;

    Terminator::PerformAll();

    return 0;
  }
  catch (std::bad_alloc) {
    std::cerr << "Memory allocation failed, LiLFeS terminated." << std::endl;
    throw;
  }
  catch (std::exception &e) {
    std::cerr << "Fatal error: " << e.what() << std::endl;
    throw;
  }
}

