/*
 * $Id: maxent.h,v 1.3 2005/06/21 10:50:49 yusuke Exp $
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 */

#ifndef __MAXENT_H_
#define __MAXENT_H_

#include <string>
#include <vector>
#include <list>
#include <map>
#include <algorithm>
#include <iostream>
#include <string>
#include <cassert>

typedef struct ME_Model_Data
{
  char * label;
  char * feature;
  double weight;
} ME_Model_Data;

//#include "blmvm.h"

struct ME_Sample
{
  std::string label;
  std::list<std::string> features;
};

class ME_Model
{
public:

  int train(const std::vector<ME_Sample> & train,
            const int cutoff = 3, const double sigma = 0, const double widthfactor = 0);
  std::string classify(const ME_Sample & nbs, std::vector<double> * membp = NULL) const;
  bool load_from_file(const std::string & filename);
  bool save_to_file(const std::string & filename) const;
  int num_classes() const { return _num_classes; }
  std::string get_class_name(int i) const { return _label_bag.Str(i); }
  void get_features(std::list< std::pair< std::pair<std::string, std::string>, double> > & fl);
  void setparam_heldout(const int h, const int n) { _nheldout = h; _early_stopping_n = n; };
  bool load_from_array(const ME_Model_Data data[]);

  ME_Model() {
    _nheldout = 0;
    _early_stopping_n = 0;
  }

private:  

  struct Sample {
    int label;
    std::list<int> positive_features;
  };

  struct ME_Feature
  {
    int label;
    int feature;
    bool operator<(const ME_Feature & x) const {
      if (label != x.label) return label < x.label;
      return feature < x.feature;
    }
    ME_Feature(const int l, const int f) : label(l), feature(f) {};
  };

  struct ME_FeatureBag
  {
    std::map<ME_Feature, int> mef2id;
    std::vector<ME_Feature> id2mef;
    int Put(const ME_Feature & i) {
      std::map<ME_Feature, int>::const_iterator j = mef2id.find(i);
      if (j == mef2id.end()) {
        int id = id2mef.size();
        id2mef.push_back(i);
        mef2id[i] = id;
        return id;
      }
      return j->second;
    }
    int Id(const ME_Feature & i) const {
      std::map<ME_Feature, int>::const_iterator j = mef2id.find(i);
      if (j == mef2id.end()) {
        return -1;
      }
      return j->second;
    }
    ME_Feature Feature(int id) const {
      assert(id >= 0 && id < (int)id2mef.size());
      return id2mef[id];
    }
    int Size() const {
      return id2mef.size();
    }
  };

  struct StringBag
  {
    std::map<std::string, int> str2id;
    std::vector<std::string> id2str;
    int Put(const std::string & i) {
      std::map<std::string, int>::const_iterator j = str2id.find(i);
      if (j == str2id.end()) {
        int id = id2str.size();
        id2str.push_back(i);
        str2id[i] = id;
        return id;
      }
      return j->second;
    }
    int Id(const std::string & i) const {
      std::map<std::string, int>::const_iterator j = str2id.find(i);
      if (j == str2id.end()) {
        return -1;
      }
      return j->second;
    }
    std::string Str(const int id) const {
      assert(id >= 0 && id < (int)id2str.size());
      return id2str[id];
    }
    int Size() const {
      return id2str.size();
    }
  };

  
  StringBag _label_bag;
  StringBag _featurename_bag;
  double _sigma; // Gaussian prior
  double _inequality_width;
  std::vector<double> _vl;  // vector of lambda
  std::vector<double> _va;  // vector of alpha (for inequality ME)
  std::vector<double> _vb;  // vector of beta  (for inequality ME)
  ME_FeatureBag _fb;
  int _num_classes;
  std::vector<double> _vee;  // empirical expectation
  std::vector<double> _vme;  // empirical expectation
  std::vector< std::vector< std::list<int> > > _sample2feature;
  std::vector< Sample > _train;
  std::vector< Sample > _heldout;
  double _train_error;   // current error rate on the training data
  double _heldout_error; // current error rate on the heldout data
  int _nheldout;
  int _early_stopping_n;
  std::vector<double> _vhlogl;

  double heldout_likelihood();
  void conditional_probability(const Sample & nbs, std::vector<double> & membp) const;
  int make_feature_bag(const int cutoff);
  int classify(const Sample & nbs, std::vector<double> & membp) const;
  double update_model_expectation();
  int perform_LMVM();

  // BLMVM
  //  int BLMVMComputeFunctionGradient(BLMVM blmvm, BLMVMVec X,double *f,BLMVMVec G);
  //  int BLMVMComputeBounds(BLMVM blmvm, BLMVMVec XL, BLMVMVec XU);
  int BLMVMSolve(double *x, int n);
  int BLMVMFunctionGradient(double *x, double *f, double *g, int n);
  int BLMVMLowerAndUpperBounds(double *xl,double *xu,int n);
  //  int Solve_BLMVM(BLMVM blmvm, BLMVMVec X);
  
};


#endif


/*
 * $Log: maxent.h,v $
 * Revision 1.3  2005/06/21 10:50:49  yusuke
 *
 * copyright 񤭤ޤ
 *
 * Revision 1.2  2005/06/16 07:25:33  yusuke
 *
 * integrate-up 硥
 * version 2.0
 *
 * Revision 1.1.2.1  2005/06/09 10:31:25  yusuke
 *
 * tagger 򤳤˰ܤޤ
 *
 * Revision 1.1.2.1  2005/06/05 10:12:19  ninomi
 * upmayz/srcβ˰ܤϤޤ
 *
 * Revision 1.2  2004/11/16 09:13:54  ninomi
 * Artistic LicenseƬˤĤޤ
 *
 * Revision 1.1  2004/08/13 07:00:15  tsuruoka
 * move maxent.h into include/up/
 *
 * Revision 1.1  2004/07/27 08:21:28  tsuruoka
 * add a part-of-speech tagger
 *
 * Revision 1.18  2004/07/22 08:34:45  tsuruoka
 * modify _sample2feature[]
 *
 * Revision 1.17  2004/07/21 16:33:01  tsuruoka
 * remove some comments
 *
 */
