/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file outputxml.cc
 *  @version Time-stamp: <2005-07-08 13:56:27 yusuke>
 *  For XML-style output
 *
 **********************************************************************/

#include "config_mayz.h"
#include <liblilfes/lilfes.h>
#include <liblilfes/builtin.h>
#include <liblilfes/stream.h>
#include <liblilfes/utility.h>

using namespace lilfes;
using namespace std;

//////////////////////////////////////////////////////////////////////

static const char* OUTPUTXML_MODULE_NAME = "enju/outputxml";

static const module* outputxml_module = NULL;
static const type* t_xml_word_or_phrase = NULL;
static const type* t_xml_word = NULL;
static const type* t_xml_phrase_unary = NULL;
static const type* t_xml_phrase_binary = NULL;
static const type* t_xml_arg = NULL;
static const feature* f_xml_phrase_id = NULL;
static const feature* f_xml_phrase_sym = NULL;
static const feature* f_xml_input_word = NULL;
static const feature* f_xml_input_pos = NULL;
static const feature* f_xml_word_base = NULL;
static const feature* f_xml_arg_list = NULL;
static const feature* f_xml_arg_label = NULL;
static const feature* f_xml_arg_id = NULL;
static const feature* f_xml_head_id = NULL;
static const feature* f_xml_phrase_udtr = NULL;
static const feature* f_xml_phrase_ldtr = NULL;
static const feature* f_xml_phrase_rdtr = NULL;

static bool initialize_types( machine& mach ) {
  outputxml_module = load_module( mach, "outputxml module", OUTPUTXML_MODULE_NAME );
  if ( ! outputxml_module ) return false;
  t_xml_word_or_phrase = module::UserModule()->Search( "xml_word_or_phrase" );
  t_xml_word = module::UserModule()->Search( "xml_word" );
  t_xml_phrase_unary = module::UserModule()->Search( "xml_phrase_unary" );
  t_xml_phrase_binary = module::UserModule()->Search( "xml_phrase_binary" );
  t_xml_arg = module::UserModule()->Search( "xml_arg" );
  f_xml_phrase_id = module::UserModule()->SearchFeature( "XML_PHRASE_ID\\" );
  f_xml_phrase_sym = module::UserModule()->SearchFeature( "XML_PHRASE_SYM\\" );
  f_xml_input_word = module::UserModule()->SearchFeature( "XML_INPUT_WORD\\" );
  f_xml_input_pos = module::UserModule()->SearchFeature( "XML_INPUT_POS\\" );
  f_xml_word_base = module::UserModule()->SearchFeature( "XML_WORD_BASE\\" );
  f_xml_arg_list = module::UserModule()->SearchFeature( "XML_ARG_LIST\\" );
  f_xml_arg_label = module::UserModule()->SearchFeature( "XML_ARG_LABEL\\" );
  f_xml_arg_id = module::UserModule()->SearchFeature( "XML_ARG_ID\\" );
  f_xml_head_id = module::UserModule()->SearchFeature( "XML_HEAD_ID\\" );
  f_xml_phrase_udtr = module::UserModule()->SearchFeature( "XML_PHRASE_UDTR\\" );
  f_xml_phrase_ldtr = module::UserModule()->SearchFeature( "XML_PHRASE_LDTR\\" );
  f_xml_phrase_rdtr = module::UserModule()->SearchFeature( "XML_PHRASE_RDTR\\" );
  return t_xml_word_or_phrase
    && t_xml_word
    && t_xml_phrase_unary
    && t_xml_phrase_binary
    && t_xml_arg
    && f_xml_phrase_id
    && f_xml_phrase_sym
    && f_xml_input_word
    && f_xml_input_pos
    && f_xml_word_base
    && f_xml_arg_list
    && f_xml_arg_label
    && f_xml_arg_id
    && f_xml_head_id
    && f_xml_phrase_udtr
    && f_xml_phrase_ldtr
    && f_xml_phrase_rdtr;
}

//////////////////////////////////////////////////////////////////////

struct xml_tag {
public:
  typedef pair< string::size_type, string::size_type > position;
  string tag;
  vector< pair< string, string > > attrs;
  position pos;

  xml_tag( const string& t, position p ) : tag( t ), pos( p ) {}
};

class output_xml_error : public exception {
private:
  const char* msg;
public:
  output_xml_error( const char* m ) : msg( m ) {}
  ~output_xml_error() throw() {}
  const char* what() const throw() { return msg; }
};

static string inttostr( int i ) {
  ostringstream s;
  s << i;
  return s.str();
}

//////////////////////////////////////////////////////////////////////

static string encode_xml_string( const string& str ) {
  string out;
  string::size_type spos = 0;
  string::size_type epos = 0;
  while ( ( epos = str.find_first_of( "&<>\"\'", spos ) ) != string::npos ) {
    out += str.substr( spos, epos - spos );
    switch ( str[ epos ] ) {
    case '&':
      out += "&amp;";
      break;
    case '<':
      out += "&lt;";
      break;
    case '>':
      out += "&gt;";
      break;
    case '"':
      out += "&quot;";
      break;
    case '\'':
      out += "&apos;";
      break;
    default:
      ;
    }
    spos = epos + 1;
  }
  out += str.substr( spos );
  return out;
}

//////////////////////////////////////////////////////////////////////

static xml_tag::position make_tag_stack( machine& mach, FSP xml, const string& text, xml_tag::position position, vector< xml_tag >& tag_stack ) {
  static bool initialized = initialize_types( mach );
  if ( ! initialized ) {
    throw output_xml_error( "output_xml: loading lilfes module failed" );
  }
  const type* t = xml.GetType();
  if ( ! t->IsSubType( t_xml_word_or_phrase ) ) {
    throw output_xml_error( "output_xml: input must be xml_word_or_phrase" );
  }
  int phrase_id = 0;
  if ( ! lilfes_to_c< int >().convert( mach, xml.Follow( f_xml_phrase_id ), phrase_id ) ) {
    throw output_xml_error( "output_xml: XML_PHRASE_ID is not filled" );
  }
  if ( t == t_xml_word ) {
    // terminal
    string word;
    if ( ! lilfes_to_c< string >().convert( mach, xml.Follow( f_xml_input_word ), word ) ) {
      throw output_xml_error( "output_xml: XML_INPUT_WORD is not filled" );
    }
    string::size_type spos = text.rfind( word, position.second );
    string::size_type len = word.size();
    if ( word == "``" || word == "''" ) {
      // double quotation
      // (needs special handling because of the tokenizer modifying an input text)
      string::size_type dpos = text.rfind( "\"", position.second );
      if ( spos == string::npos ||
           ( dpos != string::npos && dpos > spos ) ) {
        spos = dpos;
        len = 1;
      }
    }
    if ( spos == string::npos ) {
      throw output_xml_error( "output_xml: input text and xml tree are inconsistent" );
    }
    if ( spos < position.first ) {
      throw output_xml_error( "output_xml: input text and xml tree are inconsistent" );
    }
    string pos;
    if ( ! lilfes_to_c< string >().convert( mach, xml.Follow( f_xml_input_pos ), pos ) ) {
      throw output_xml_error( "output_xml: XML_INPUT_POS is not filled" );
    }
    string base;
    if ( ! lilfes_to_c< string >().convert( mach, xml.Follow( f_xml_word_base ), base ) ) {
      throw output_xml_error( "output_xml: XML_WORD_BASE is not filled" );
    }
    vector< FSP > args;
    if ( ! lilfes_to_c< vector< FSP > >().convert( mach, xml.Follow( f_xml_arg_list ), args ) ) {
      throw output_xml_error( "output_xml: argument list in a xml tree is inconsistent" );
    }
    xml_tag word_tag( "word", make_pair( spos, spos + len ) );
    word_tag.attrs.push_back( make_pair( "id", inttostr( phrase_id ) ) );
    word_tag.attrs.push_back( make_pair( "pos", pos ) );
    word_tag.attrs.push_back( make_pair( "base", base ) );
    for ( vector< FSP >::iterator it = args.begin();
          it != args.end();
          ++it ) {
      if ( it->GetType() != t_xml_arg ) {
        throw output_xml_error( "output_xml: argument list must consist of 'xml_arg'" );
      }
      string arg_label;
      if ( ! lilfes_to_c< string >().convert( mach, it->Follow( f_xml_arg_label ), arg_label ) ) {
        throw output_xml_error( "output_xml: argument label in a xml tree is inconsistent" );
      }
      int arg_id = 0;
      if ( ! lilfes_to_c< int >().convert( mach, it->Follow( f_xml_arg_id ), arg_id ) ) {
        throw output_xml_error( "output_xml: argument id in a xml tree is inconsistent" );
      }
      word_tag.attrs.push_back( make_pair( arg_label, inttostr( arg_id ) ) );
    }
    tag_stack.push_back( word_tag );
    return word_tag.pos;
  }
  // nonterminal
  string sym;
  if ( ! lilfes_to_c< string >().convert( mach, xml.Follow( f_xml_phrase_sym ), sym ) ) {
    throw output_xml_error( "output_xml: XML_PHRASE_SYM is not filled" );
  }
  int head_id = 0;
  if ( ! lilfes_to_c< int >().convert( mach, xml.Follow( f_xml_head_id ), head_id ) ) {
    throw output_xml_error( "output_xml: XML_HEAD_ID is not filled" );
  }
  if ( t == t_xml_phrase_unary ) {
    FSP udtr = xml.Follow( f_xml_phrase_udtr );
    xml_tag::position upos = make_tag_stack( mach, udtr, text, position, tag_stack );
    xml_tag phrase_tag( "phrase", upos );
    phrase_tag.attrs.push_back( make_pair( "id", inttostr( phrase_id ) ) );
    phrase_tag.attrs.push_back( make_pair( "cat", sym ) );
    phrase_tag.attrs.push_back( make_pair( "head", inttostr( head_id ) ) );
    tag_stack.push_back( phrase_tag );
    return upos;
  } else if ( t == t_xml_phrase_binary ) {
    FSP rdtr = xml.Follow( f_xml_phrase_rdtr );
    xml_tag::position rpos = make_tag_stack( mach, rdtr, text, position, tag_stack );
    FSP ldtr = xml.Follow( f_xml_phrase_ldtr );
    xml_tag::position lpos = make_tag_stack( mach, ldtr, text, make_pair( position.first, rpos.first ), tag_stack );
    xml_tag phrase_tag( "phrase", make_pair( lpos.first, rpos.second ) );
    phrase_tag.attrs.push_back( make_pair( "id", inttostr( phrase_id ) ) );
    phrase_tag.attrs.push_back( make_pair( "cat", sym ) );
    phrase_tag.attrs.push_back( make_pair( "head", inttostr( head_id ) ) );
    tag_stack.push_back( phrase_tag );
    return phrase_tag.pos;
  } else {
    throw output_xml_error( "output_xml: xml_word_or_phrase is not filled" );
  }
}

//////////////////////////////////////////////////////////////////////

static void output_tag_stack_so(const string& text_id, const vector< xml_tag >& tag_stack, ostream& ostr) {
  for ( vector< xml_tag >::const_reverse_iterator it = tag_stack.rbegin();
        it != tag_stack.rend();
        ++it ) {
    ostr << text_id << '\t' << it->pos.first << '\t' << it->pos.second << '\t' << it->tag;
    for ( vector< pair< string, string > >::const_iterator attr_it = it->attrs.begin();
          attr_it != it->attrs.end();
          ++attr_it ) {
      ostr << ' ' << attr_it->first << "=\"" << encode_xml_string( attr_it->second ) << '\"';
    }
    ostr << '\n';
  }
}

static void output_tag_stack_xml(const string& text, const vector< xml_tag >& tag_stack, ostream& ostr) {
  string::size_type spos = 0;
  vector< vector< xml_tag >::const_reverse_iterator > end_stack;
  for ( vector< xml_tag >::const_reverse_iterator it = tag_stack.rbegin();
        it != tag_stack.rend();
        ++it ) {
    while ( ! end_stack.empty() && it->pos.first >= end_stack.back()->pos.second ) {
      ostr << encode_xml_string( text.substr( spos, end_stack.back()->pos.second - spos ) );
      spos = end_stack.back()->pos.second;
      ostr << "</" << end_stack.back()->tag << '>';
      end_stack.pop_back();
    }
    ostr << encode_xml_string( text.substr( spos, it->pos.first - spos ) );
    spos = it->pos.first;
    ostr << '<' << it->tag;
    for ( vector< pair< string, string > >::const_iterator attr_it = it->attrs.begin();
          attr_it != it->attrs.end();
          ++attr_it ) {
      ostr << ' ' << attr_it->first << "=\"" << encode_xml_string( attr_it->second ) << '\"';
    }
    ostr << '>';
    end_stack.push_back( it );
  }
  while ( ! end_stack.empty() ) {
    ostr << encode_xml_string( text.substr( spos, end_stack.back()->pos.second - spos ) );
    spos = end_stack.back()->pos.second;
    ostr << "</" << end_stack.back()->tag << '>';
    end_stack.pop_back();
  }
  ostr << encode_xml_string( text.substr( spos ) );
}

//////////////////////////////////////////////////////////////////////

static bool output_phrase_structure_so(machine& mach, FSP text_id, FSP text, FSP xml, FSP outstr) {
  if( ! text_id.IsString() ) {
    RUNERR( "arg1 of output_so/4 must be a string" );
    return false;
  }
  std::string text_id_str( text_id.ReadString() );
  if ( ! text.IsString() ) {
    RUNERR( "arg2 of output_so/4 must be a string" );
    return false;
  }
  std::string text_str( text.ReadString() );

  xml_tag::position pos = std::make_pair( 0, text_str.size() );

  lilfes_stream *s = lilfes_stream::GetStreamFSP( outstr );
  if ( s == NULL ) {
    RUNERR( "An invalid handle in the 3rd argument of 'output_xml/3'" );
    return false;
  }
  std::ostream* ostr = dynamic_cast< std::ostream* >( s->GetCStream() );
  if ( ostr == NULL ) {
    RUNERR( "Stream is not an output stream in 'output_xml/3'" );
    return false;
  }

  if ( ! (*ostr) ) return false;

  vector< xml_tag > tag_stack;
  try {
    make_tag_stack( mach, xml, text_str, pos, tag_stack );
    output_tag_stack_so( text_id_str, tag_stack, *ostr );
  } catch ( output_xml_error& e ) {
    RUNERR( e.what() );
    return false;
  }

  if ( ! (*ostr) ) return false;

  return true;
}

static bool output_phrase_structure_xml(machine& mach, FSP text, FSP xml, FSP outstr) {
  if( ! text.IsString() ) {
    RUNERR( "arg1 of output_xml/3 must be a string" );
    return false;
  }
  std::string text_str( text.ReadString() );

  xml_tag::position pos = std::make_pair(0, text_str.size());

  lilfes_stream *s = lilfes_stream::GetStreamFSP(outstr);
  if ( s == NULL ) {
    RUNERR( "An invalid handle in the 3rd argument of 'output_xml/3'" );
    return false;
  }
  std::ostream* ostr = dynamic_cast< std::ostream* >( s->GetCStream() );
  if ( ostr == NULL ) {
    RUNERR( "Stream is not an output stream in 'output_xml/3'" );
    return false;
  }

  if ( ! (*ostr) ) return false;

  vector< xml_tag > tag_stack;
  try {
    make_tag_stack( mach, xml, text_str, pos, tag_stack );
    output_tag_stack_xml( text_str, tag_stack, *ostr );
  } catch ( output_xml_error& e ) {
    RUNERR( e.what() );
    return false;
  }

  if ( ! (*ostr) ) return false;

  return true;
}

LILFES_BUILTIN_PRED_4(output_phrase_structure_so, output_phrase_structure_so);
LILFES_BUILTIN_PRED_3(output_phrase_structure_xml, output_phrase_structure_xml);

