/*
 *
 * $Id: Edge.h,v 1.3 2005/06/21 10:17:12 yusuke Exp $
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 */

#ifndef __Edge_h
#define __Edge_h

#include <liblilfes/structur.h>
#include <utility>
#include <vector>
#include "Fs.h"

namespace up {

////////////////////////////////////////////////////
//////
////// Edge
//////

    typedef lilfes::uint32 eserial;
    typedef lilfes::uint32 lserial;
    
    class Link;

    class Edge {
    private:
        eserial serialno;
        std::pair<int, int> position; // left postion and right position
        Fs* sign;
        std::vector<Link*> links;
    public:
        Edge(eserial __serial_no, std::pair<int, int> __position, Fs* __sign)
                : serialno(__serial_no),
                  position(__position),
                  sign(__sign) {}
        ~Edge() { if(sign) delete sign; }
    
        Fs* getSign() { return sign; }
        const Fs* getSign() const { return sign; }
        void setSign(Fs* s) { sign = s; }

        std::pair<int, int> getPosition() { return position; }
        void setPosition(std::pair<int, int> p) { position = p; }
        
        int getLeftPosition() const { return position.first; }
        void setLeftPosition(int p) { position.first = p; }
        
        int getRightPosition() const { return position.second; }
        void setRightPosition(int p) { position.second = p; }
        
        eserial getSerialNo() const { return serialno; }
        void setSerialNo(eserial sn) { serialno = sn; }
        
        std::vector<Link*>& getLinks() { return links; }
        void addLink(Link*);
    };

    class EdgeManager {
    private:
        std::vector<Edge*> serialtable;
    public:
        EdgeManager() {}
        ~EdgeManager() {
            std::vector<Edge*>::iterator it = serialtable.begin(), last = serialtable.end();
            for( ; it != last ; it++)
                delete(*it);
        }
	Edge* makeEdge(std::pair<int, int> p, Fs* x) {
	    Edge* e = new Edge((eserial) serialtable.size(), p, x);
            serialtable.push_back(e);
	    return e;
	}
        Edge* getEdge(eserial e) {
	    if(e >= serialtable.size())
		return (Edge*)0;
	    else
		return serialtable[e];
	}
        eserial getSerialCount() {return (eserial) serialtable.size(); }
        void clear() {
            std::vector<Edge*>::iterator it = serialtable.begin(), last = serialtable.end();
            for( ; it != last ; it++)
                delete(*it);
            serialtable.clear();
        }
    };

////////////////////////////////////////////////////
//////
////// Link
//////
    class Link {
    private:
        lserial serialno;
	std::pair<int, int> position;
        Fs* sign_plus;
        eserial my_edge_serialno;
    public:
        Link(lserial __serialno, std::pair<int, int> __position, Fs* __sign_plus)
                : serialno(__serialno), position(__position), sign_plus(__sign_plus) {}
        virtual ~Link() {
            if(sign_plus) delete sign_plus;
        };

	std::pair<int, int> getPosition() const { return position; }
	void setPosition(std::pair<int, int> p) { position = p; }
	
        int getLeftPosition() const { return position.first; }
        void setLeftPosition(int p) { position.first = p; }
        
        int getRightPosition() const { return position.second; }
        void setRightPosition(int p) { position.second = p; }

        lserial getSerialNo() const { return serialno; }
        void setSerialNo(lserial sn) { serialno = sn; }
        
        Fs* getSignPlus() const { return sign_plus; }
        void setSignPlus(Fs* s) { sign_plus = s; }
        
        eserial getMyEdgeSerialNo() { return my_edge_serialno; }
        void setMyEdgeSerialNo(eserial sn) { my_edge_serialno = sn; }
    };

    class LinkBinary : public Link {
    private:
        Edge* l_dtr;           // the edge of my left daughter
        Edge* r_dtr;           // the edge of my right daughter
        Fs* applied_schema;  // the name of the applied schema
    public:
        LinkBinary(lserial __serialno,
                   std::pair<int, int> __position,
                   Fs* __sign_plus,
                   Edge* __l_dtr,
                   Edge* __r_dtr,
                   Fs* __applied_schema)
                : Link(__serialno, __position, __sign_plus),
                  l_dtr(__l_dtr),
                  r_dtr(__r_dtr),
                  applied_schema(__applied_schema) {}
        ~LinkBinary() { if(applied_schema) delete applied_schema; }

        Edge* getLDtr() { return l_dtr; }
        Edge* getRDtr() { return r_dtr; }
        std::pair<int, int> getLCell() { return l_dtr->getPosition(); }
        std::pair<int, int> getRCell() { return r_dtr->getPosition(); }
        Fs* getAppliedSchema() { return applied_schema; }
    };

    class LinkUnary : public Link {
    private:
        Edge* u_dtr;           // the edge of my unary daughter
        Fs* applied_schema;  // the name of an applied schema
    public:
        LinkUnary(lserial __serialno,
                  std::pair<int, int> __position,
                  Fs* __sign_plus,
                  Edge* __u_dtr,
                  Fs* __applied_schema)
                : Link(__serialno, __position, __sign_plus),
                  u_dtr(__u_dtr),
                  applied_schema(__applied_schema) {}
        ~LinkUnary() { if(applied_schema) delete applied_schema; }
        Edge* getUDtr()         { return u_dtr; }
        std::pair<int, int> getUCell() { return u_dtr->getPosition(); }
        Fs* getAppliedSchema()    { return applied_schema; }
    };

    class LinkTerminal : public Link {
    private:
        Fs* lex_name;            // 3rd argument of 'ext_lexical_entry/4' if this is a terminal edge
    public:    
        LinkTerminal(lserial __serialno,
                     std::pair<int, int> __position,
                     Fs* __sign_plus,
                     Fs* __lex_name)
                : Link(__serialno, __position, __sign_plus),
                  lex_name(__lex_name) {}
        ~LinkTerminal() { if(lex_name) delete lex_name; }
        Fs* getLexName()          { return lex_name; }
    };

    class LinkManager {
    private:
        std::vector<Link*> serialtable;
    public:
        LinkManager() {}
        ~LinkManager() {
            std::vector<Link*>::iterator it = serialtable.begin(), last = serialtable.end();
            for( ; it != last ; it++)
                delete (*it);
        }
	LinkBinary* makeLinkBinary(std::pair<int, int> position,
                                   Fs* sign_plus, Edge* l_dtr, Edge* r_dtr,
                                   Fs* applied_schema) {
	    LinkBinary* i = new LinkBinary((lserial) serialtable.size(),
                                           position, sign_plus, l_dtr, r_dtr,
                                           applied_schema);
            serialtable.push_back(i);
	    return i;
        }
	LinkUnary* makeLinkUnary(std::pair<int, int> position,
                                 Fs* sign_plus, Edge* u_dtr,
                                 Fs* applied_schema) {
	    LinkUnary* i = new LinkUnary((lserial) serialtable.size(),
                                         position, sign_plus, u_dtr,
                                         applied_schema);
	    serialtable.push_back(i);
	    return i;
	}
	LinkTerminal* makeLinkTerminal(std::pair<int, int> position,
                                       Fs* sign_plus,
                                       Fs* lex_name) {
	    LinkTerminal* i = new LinkTerminal((lserial) serialtable.size(),
                                               position, sign_plus, lex_name);
	    serialtable.push_back(i);
	    return i;
	}
        Link* getLink(lserial e) { if( e >= serialtable.size()) return (Link*)0; else return serialtable[e]; };
        lserial getSerialCount() {return (lserial) serialtable.size(); };
        void clear() {
            std::vector<Link*>::iterator it = serialtable.begin(), last = serialtable.end();
            for( ; it != last ; it++)
                delete (*it);
            serialtable.clear();
        }
    };
    
} // end of namespace up

#endif // __Edge_h

/*
 * $Log: Edge.h,v $
 * Revision 1.3  2005/06/21 10:17:12  yusuke
 *
 * copyright 񤭤ޤ
 *
 * Revision 1.2  2005/06/16 07:25:36  yusuke
 *
 * integrate-up 硥
 * version 2.0
 *
 * Revision 1.1.2.2  2005/06/05 14:10:48  ninomi
 * ѥ뤬Ȥ褦ˤʤޤ
 *
 * Revision 1.1.2.1  2005/06/05 10:12:18  ninomi
 * upmayz/srcβ˰ܤϤޤ
 *
 * Revision 1.13  2005/04/13 06:29:30  tsuruoka
 * add files for best-first parsing
 *
 * Revision 1.12  2004/11/16 09:13:54  ninomi
 * Artistic LicenseƬˤĤޤ
 *
 * Revision 1.11  2004/06/30 07:41:48  ninomi
 * up 0.2˰ܹԤ뤿0.1θŤե
 *
 */
