/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file ForestMaker.cc
 *  @version Time-stamp: <2005-06-21 19:06:48 yusuke>
 *  Implementation of forestmaker
 *
 **********************************************************************/

#include "AmisModel.h"
#include "ForestMaker.h"
#include "Initializer.h"

namespace mayz {

  using namespace std;
  using namespace lilfes;
  using namespace up;

  bool ForestMaker::init( machine* m, Grammar* g ) {
    if(! ParserCky::init(m, g)) return false;

    module* forestmake_module = load_module( *m, "forestmake module", "mayz/forestmake" );
    if ( ! forestmake_module ) {
      RUNERR( "module nod found: mayz/forestmake" );
      return false;
    }
    if ( ! Initializer::initialize( m ) ) {
      RUNERR( "initialization of forest_maker failed" );
      return false;
    }
    convert_derivation = prepare_proc( module::UserModule(), "convert_derivation", 2 );
    if ( ! convert_derivation ) {
      RUNERR( "predicate not defined: convert_derivation/2 (maybe inconsistency of forestmaker version" );
      return false;
    }
    extract_terminal_event = prepare_proc( module::UserModule(), "extract_terminal_event", 6 );
    extract_unary_event = prepare_proc( module::UserModule(), "extract_unary_event", 7 );
    extract_binary_event = prepare_proc( module::UserModule(), "extract_binary_event", 8 );
    extract_root_event = prepare_proc( module::UserModule(), "extract_root_event", 4 );

    extract_terminal_event_feature_value = prepare_proc( module::UserModule(), "extract_terminal_event_feature_value", 7 );
    extract_unary_event_feature_value = prepare_proc( module::UserModule(), "extract_unary_event_feature_value", 8 );
    extract_binary_event_feature_value = prepare_proc( module::UserModule(), "extract_binary_event_feature_value", 9 );
    extract_root_event_feature_value = prepare_proc( module::UserModule(), "extract_root_event_feature_value", 5 );

    reference_prob_terminal = prepare_proc( module::UserModule(), "reference_prob_terminal", 5 );
    reference_prob_unary = prepare_proc( module::UserModule(), "reference_prob_unary", 6 );
    reference_prob_binary = prepare_proc( module::UserModule(), "reference_prob_binary", 7 );
    reference_prob_root = prepare_proc( module::UserModule(), "reference_prob_root", 3 );

    return true;
  }

  //////////////////////////////////////////////////////////////////////

  bool ForestMaker::makeWordLattice( Derivation derivation, WordLattice& wordlattice ) {
    //core_p trailp = mach->SetTrailPoint( NULL );
    FSP wl( mach );
    if ( ! grammar->sentence2WordLattice( derivation, wl ) ) {
      //mach->TrailBack( trailp );
      return false;
    }
    while( wl.GetType()->IsSubType(cons) ) {
      wordlattice.push_back(fspm->newExtent(wl.Follow(hd)));
      wl = wl.Follow(tl);
    }
    //mach->TrailBack( trailp );
    return true;
  }

  bool ForestMaker::derivationToParseTree( Derivation derivation, ParseTree parse_tree ) {
    vector< FSP > args( 2 );
    args[ 0 ] = derivation;
    args[ 1 ] = parse_tree;
    core_p cutp = mach->GetCutPoint();
    if ( ! call_proc( *mach, convert_derivation, args ) ) {
      return false;
    }
    mach->DoCut( cutp );
    return true;
  }

  //////////////////////////////////////////////////////////////////////

  bool ForestMaker::outputTerminalEvents( FSP lexname, FSP sign, FSP sign_plus, ostream& output_file ) {
    if ( extract_terminal_event ) {
      vector< FSP > args( 6 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = lexname;
      args[ 3 ] = sign;
      args[ 4 ] = sign_plus;
      args[ 5 ] = FSP( mach );    // event
      vector< bool > flags( 6, false );
      flags[ 1 ] = true;
      flags[ 5 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_terminal_event, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_terminal_event/6 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "6th argument of extract_terminal_event/6 must be list of strings" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, event_name );
        output_file << ' ' << event_name;
      }
    }
    if ( extract_terminal_event_feature_value ) {
      vector< FSP > args( 7 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = lexname;
      args[ 3 ] = sign;
      args[ 4 ] = sign_plus;
      args[ 5 ] = FSP( mach );    // event
      args[ 6 ] = FSP( mach );    // feature value
      vector< bool > flags( 7, false );
      flags[ 1 ] = true;
      flags[ 5 ] = true;
      flags[ 6 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_terminal_event_feature_value, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        float feature_value;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_terminal_event_feature_value/7 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "6th argument of extract_terminal_event_feature_value/7 must be list of strings" );
          return false;
        }
        if ( ! lilfes_to_c< float >().convert( *mach, (*it)[ 2 ], feature_value ) ) {
          RUNWARN( "7th argument of extract_terminal_event_feature_value/7 must be float" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, feature_value, event_name );
        output_file << ' ' << event_name;
      }
    }
    return true;
  }

  bool ForestMaker::outputUnaryEvents( FSP schema, FSP dtr, FSP mother, FSP sign_plus, ostream& output_file ) {
    if ( extract_unary_event ) {
      vector< FSP > args( 7 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = schema;
      args[ 3 ] = dtr;
      args[ 4 ] = mother;
      args[ 5 ] = sign_plus;
      args[ 6 ] = FSP( mach );    // ME event
      vector< bool > flags( 7, false );
      flags[ 1 ] = true;
      flags[ 6 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_unary_event, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_unary_event/7 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "7th argument of extract_unary_event/7 must be list of strings" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, event_name );
        output_file << ' ' << event_name;
      }
    }
    if ( extract_unary_event_feature_value ) {
      vector< FSP > args( 8 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = schema;
      args[ 3 ] = dtr;
      args[ 4 ] = mother;
      args[ 5 ] = sign_plus;
      args[ 6 ] = FSP( mach );    // ME event
      args[ 7 ] = FSP( mach );    // feature value
      vector< bool > flags( 8, false );
      flags[ 1 ] = true;
      flags[ 6 ] = true;
      flags[ 7 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_unary_event_feature_value, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        float feature_value;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_unary_event_feature_value/8 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "7th argument of extract_unary_event_feature_value/8 must be list of strings" );
          return false;
        }
        if ( ! lilfes_to_c< float >().convert( *mach, (*it)[ 2 ], feature_value ) ) {
          RUNWARN( "8th argument of extract_unary_event_feature_value/8 must be float" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, feature_value, event_name );
        output_file << ' ' << event_name;
      }
    }
    return true;
  }

  bool ForestMaker::outputBinaryEvents( FSP schema, FSP ldtr, FSP rdtr, FSP mother, FSP sign_plus, ostream& output_file ) {
    if ( extract_binary_event ) {
      vector< FSP > args( 8 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = schema;
      args[ 3 ] = ldtr;
      args[ 4 ] = rdtr;
      args[ 5 ] = mother;
      args[ 6 ] = sign_plus;
      args[ 7 ] = FSP( mach );    // ME event
      vector< bool > flags( 8, false );
      flags[ 1 ] = true;
      flags[ 7 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_binary_event, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_binary_event/8 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "8th argument of extract_binary_event/8 must be list of strings" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, event_name );
        output_file << ' ' << event_name;
      }
    }
    if ( extract_binary_event_feature_value ) {
      vector< FSP > args( 9 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = schema;
      args[ 3 ] = ldtr;
      args[ 4 ] = rdtr;
      args[ 5 ] = mother;
      args[ 6 ] = sign_plus;
      args[ 7 ] = FSP( mach );    // ME event
      args[ 8 ] = FSP( mach );    // feature value
      vector< bool > flags( 9, false );
      flags[ 1 ] = true;
      flags[ 7 ] = true;
      flags[ 8 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_binary_event_feature_value, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        float feature_value;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_binary_event_feature_value/9 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "8th argument of extract_binary_event_feature_value/9 must be list of strings" );
          return false;
        }
        if ( ! lilfes_to_c< float >().convert( *mach, (*it)[ 2 ], feature_value ) ) {
          RUNWARN( "9th argument of extract_binary_event_feature_value/9 must be list of float" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, feature_value, event_name );
        output_file << ' ' << event_name;
      }
    }
    return true;
  }

  bool ForestMaker::outputRootEvents( FSP sign, ostream& output_file ) {
    if ( extract_root_event ) {
      vector< FSP > args( 4 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = sign;
      args[ 3 ] = FSP( mach );    // event
      vector< bool > flags( 4, false );
      flags[ 1 ] = true;
      flags[ 3 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_root_event, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_root_event/4 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "4th argument of extract_root_event/4 must be list of strings" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, event_name );
        output_file << ' ' << event_name;
      }
    }
    if ( extract_root_event_feature_value ) {
      vector< FSP > args( 5 );
      args[ 0 ] = FSP( mach, model_name.c_str() );
      args[ 1 ] = FSP( mach );    // category
      args[ 2 ] = sign;
      args[ 3 ] = FSP( mach );    // event
      args[ 4 ] = FSP( mach );    // feature value
      vector< bool > flags( 5, false );
      flags[ 1 ] = true;
      flags[ 3 ] = true;
      flags[ 4 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, extract_root_event_feature_value, args, flags, results );
      for ( list< vector< FSP > >::const_iterator it = results.begin();
            it != results.end();
            ++it ) {
        string category;
        vector< string > event;
        float feature_value;
        if ( ! lilfes_to_c< string >().convert( *mach, (*it)[ 0 ], category ) ) {
          RUNWARN( "2nd argument of extract_root_event_feature_value/5 must be string" );
          return false;
        }
        if ( ! lilfes_to_c< vector< string > >().convert( *mach, (*it)[ 1 ], event ) ) {
          RUNWARN( "4th argument of extract_root_event_feature_value/5 must be list of strings" );
          return false;
        }
        if ( ! lilfes_to_c< float >().convert( *mach, (*it)[ 2 ], feature_value ) ) {
          RUNWARN( "5th argument of extract_root_event_feature_value/5 must be list of strings" );
          return false;
        }
        string event_name;
        AmisModel::encode( category, event, feature_value, event_name );
        output_file << ' ' << event_name;
      }
    }
    return true;
  }

  //////////////////////////////////////////////////////////////////////

  bool ForestMaker::terminalReference( FSP lexname, FSP sign, FSP sign_plus, double& prob ) {
    if ( ! reference_prob_terminal ) {
      RUNWARN( "predicate not defined: 'reference_prob_terminal/5'" );
      return false;
    }
    vector< FSP > args( 5 );
    args[ 0 ] = FSP( mach, model_name.c_str() );
    args[ 1 ] = lexname;
    args[ 2 ] = sign;
    args[ 3 ] = sign_plus;
    args[ 4 ] = FSP( mach );  // reference probability
    core_p cutp = mach->GetCutPoint();
    if ( ! call_proc( *mach, reference_prob_terminal, args ) ) {
      RUNWARN( "execution of reference_prob_terminal/5 failed" );
      return false;
    }
    mach->DoCut( cutp );
    if ( ! lilfes_to_c< double >().convert( *mach, args[ 4 ], prob ) ) {
      RUNWARN( "5th argument of reference_prob_terminal/5 must be float" );
      return false;
    }
    return true;
  }
  bool ForestMaker::unaryReference( FSP schema_name, FSP dtr, FSP mother, FSP sign_plus, double& prob ) {
    if ( ! reference_prob_unary ) {
      RUNWARN( "predicate not defined: 'reference_prob_unary/6'" );
      return false;
    }
    vector< FSP > args( 6 );
    args[ 0 ] = FSP( mach, model_name.c_str() );
    args[ 1 ] = schema_name;
    args[ 2 ] = dtr;
    args[ 3 ] = mother;
    args[ 4 ] = sign_plus;
    args[ 5 ] = FSP( mach );  // reference probability
    core_p cutp = mach->GetCutPoint();
    if ( ! call_proc( *mach, reference_prob_unary, args ) ) {
      RUNWARN( "execution of reference_prob_unary/6 failed" );
      return false;
    }
    mach->DoCut( cutp );
    if ( ! lilfes_to_c< double >().convert( *mach, args[ 5 ], prob ) ) {
      RUNWARN( "6th argument of reference_prob_unary/6 must be float" );
      return false;
    }
    return true;
  }
  bool ForestMaker::binaryReference( FSP schema_name, FSP left_dtr, FSP right_dtr, FSP mother, FSP sign_plus, double& prob ) {
    if ( ! reference_prob_binary ) {
      RUNWARN( "predicate not defined: 'reference_prob_binary/7'" );
      return false;
    }
    vector< FSP > args( 7 );
    args[ 0 ] = FSP( mach, model_name.c_str() );
    args[ 1 ] = schema_name;
    args[ 2 ] = left_dtr;
    args[ 3 ] = right_dtr;
    args[ 4 ] = mother;
    args[ 5 ] = sign_plus;
    args[ 6 ] = FSP( mach );  // reference probability
    core_p cutp = mach->GetCutPoint();
    if ( ! call_proc( *mach, reference_prob_binary, args ) ) {
      RUNWARN( "execution of reference_prob_binary/7 failed" );
      return false;
    }
    mach->DoCut( cutp );
    if ( ! lilfes_to_c< double >().convert( *mach, args[ 6 ], prob ) ) {
      RUNWARN( "7th argument of reference_prob_binary/7 must be float" );
      return false;
    }
    return true;
  }
  bool ForestMaker::rootReference( FSP sign, double& prob ) {
    if ( ! reference_prob_root ) {
      RUNWARN( "predicate not defined: 'reference_prob_root/3'" );
      return false;
    }
    vector< FSP > args( 3 );
    args[ 0 ] = FSP( mach, model_name.c_str() );
    args[ 1 ] = sign;
    args[ 2 ] = FSP( mach );  // reference probability
    core_p cutp = mach->GetCutPoint();
    if ( ! call_proc( *mach, reference_prob_root, args ) ) {
      RUNWARN( "execution of reference_prob_root/3 failed" );
      return false;
    }
    mach->DoCut( cutp );
    if ( ! lilfes_to_c< double >().convert( *mach, args[ 2 ], prob ) ) {
      RUNWARN( "4th argument of reference_prob_root/3 must be float" );
      return false;
    }
    return true;
  }

  //////////////////////////////////////////////////////////////////////

  bool ForestMaker::outputForestConjTerm( Edge* edge, LinkTerminal* edge_link, int id ) {
    IPTrailStack iptrail( mach );
    FSP lexname( mach, edge_link->getLexName() );
    FSP sign( mach, edge->getSign() );
    FSP sign_plus( mach, edge_link->getSignPlus() );
    *output_stream << " ( term" << edge->getSerialNo() << '_' << id;
    if ( ! outputTerminalEvents( lexname, sign, sign_plus, *output_stream ) ) {
      return false;
    }
    *output_stream << " )";
    if ( reference_stream && reference_prob_terminal ) {
      double prob = 0.0;
      if ( ! terminalReference( lexname, sign, sign_plus, prob ) ) {
        return false;
      }
      *reference_stream << "term" << edge->getSerialNo() << '_' << id << '\t' << prob << '\n';
    }
    return true;
  }

  bool ForestMaker::outputForestConjUnary( Edge* edge, LinkUnary* edge_link, int id ) {
    IPTrailStack iptrail( mach );
    *output_stream << " ( u" << edge->getSerialNo() << '_' << id;
    FSP schema_name( mach, edge_link->getAppliedSchema() );
    FSP dtr( mach, edge_link->getUDtr()->getSign() );
    FSP mother( mach, edge->getSign() );
    FSP sign_plus( mach, edge_link->getSignPlus() );
    if ( ! outputUnaryEvents( schema_name, dtr, mother, sign_plus, *output_stream ) ) {
      return false;
    }
    if ( reference_stream && reference_prob_unary ) {
      double prob = 0.0;
      if ( ! unaryReference( schema_name, dtr, mother, sign_plus, prob ) ) {
        return false;
      }
      *reference_stream << 'u' << edge->getSerialNo() << '_' << id << '\t' << prob << '\n';
    }
    if ( ! outputForestDisj( edge_link->getUDtr() ) ) {
      return false;
    }
    *output_stream << " )";
    return true;
  }

  bool ForestMaker::outputForestConjBinary( Edge* edge, LinkBinary* edge_link, int id ) {
    IPTrailStack iptrail( mach );
    *output_stream << " ( b" << edge->getSerialNo() << '_' << id;
    FSP schema_name( mach, edge_link->getAppliedSchema() );
    FSP left_dtr( mach, edge_link->getLDtr()->getSign() );
    FSP right_dtr( mach, edge_link->getRDtr()->getSign() );
    FSP mother( mach, edge->getSign() );
    FSP sign_plus( mach, edge_link->getSignPlus() );
    if ( ! outputBinaryEvents( schema_name, left_dtr, right_dtr, mother, sign_plus, *output_stream ) ) {
      return false;
    }
    if ( reference_stream && reference_prob_binary ) {
      double prob = 0.0;
      if ( ! binaryReference( schema_name, left_dtr, right_dtr, mother, sign_plus, prob ) ) {
        return false;
      }
      *reference_stream << 'b' << edge->getSerialNo() << '_' << id << '\t' << prob << '\n';
    }
    if ( ( ! outputForestDisj( edge_link->getLDtr() ) )
         || ( ! outputForestDisj( edge_link->getRDtr() ) ) ) {
      return false;
    }
    *output_stream << " )";
    return true;
  }

  bool ForestMaker::outputForestDisj( Edge* edge ) {
    if ( visit_mark[ edge->getSerialNo() ] ) {
      // already visited
      *output_stream << " $node" << edge->getSerialNo();
      return true;
    } else {
      // not visited yet
      visit_mark[ edge->getSerialNo() ] = true;
      vector< Link* >& edge_links = edge->getLinks();
      *output_stream << " { node" << edge->getSerialNo();
      for ( vector< Link* >::const_iterator it = edge_links.begin();
            it != edge_links.end();
            ++it ) {
        int id = it - edge_links.begin();
        if ( typeid( **it ) == typeid( LinkBinary ) ) {
          if ( ! outputForestConjBinary( edge, dynamic_cast< LinkBinary* >( *it ), id ) ) {
            return false;
          }
        } else if ( typeid( **it ) == typeid( LinkUnary ) ) {
          if ( ! outputForestConjUnary( edge, dynamic_cast< LinkUnary* >( *it ), id ) ) {
            return false;
          }
        } else if ( typeid( **it ) == typeid( LinkTerminal ) ) {
          // terminal
          if ( ! outputForestConjTerm( edge, dynamic_cast< LinkTerminal* >( *it ), id ) ) {
            return false;
          }
        } else {
          // ??
          cerr << "Unknown edge type" << endl; ASSERT( 0 );
        }
      }
      *output_stream << " }";
      return true;
    }
  }

  bool ForestMaker::outputForestRoot() {
    visit_mark.clear();
    visit_mark.resize( getEdgeNumber() );
    std::vector<up::Edge*> &ee = getEdges(0, getSentenceLength());
    std::vector<up::Edge*>::iterator it = ee.begin(), last = ee.end();
    IPTrailStack iptrail( mach );
    *output_stream << "{ _";
    for ( ; it != last; ++it ) {
      FSP root = FSP(mach, (*it)->getSign());
      if ( grammar->rootSign( root ) ) {
        *output_stream << " ( root" << (*it)->getSerialNo();
        if ( ! outputRootEvents( root, *output_stream ) ) {
          return false;
        }
        if ( reference_stream && reference_prob_root ) {
          double prob = 0.0;
          if ( ! rootReference( root, prob ) ) {
            return false;
          }
          *reference_stream << "root" << (*it)->getSerialNo() << '\t' << prob << '\n';
        }
        if ( ! outputForestDisj( (*it) ) ) {
          return false;
        }
        *output_stream << " )";
      }
    }
    *output_stream << " }\n";
    return true;
  }

  //////////////////////////////////////////////////////////////////////

  bool ForestMaker::outputParseTreeDtrs( ParseTree parse_tree ) {
    //cerr << parse_tree.DisplayAVM();
    if ( parse_tree.isTerminal() ) {
      // terminal
      FSP lexname = parse_tree.lexName();
      FSP sign = parse_tree.edgeSign();
      FSP sign_plus = parse_tree.signPlus();
      if ( ! outputTerminalEvents( lexname, sign, sign_plus, *output_stream ) ) return false;
      if ( reference_stream && reference_prob_terminal ) {
        double prob = 0.0;
        if ( ! terminalReference( lexname, sign, sign_plus, prob ) ) {
          return false;
        }
        reference_probability *= prob;
      }
      return true;
    } else if ( parse_tree.isUnary() ) {
      // unary tree
      FSP mother = parse_tree.edgeSign();
      FSP sign_plus = parse_tree.signPlus();
      FSP schema_name = parse_tree.appliedSchema();
      ParseTree udtr_tree = parse_tree.UTree();
      FSP udtr_sign = udtr_tree.edgeSign();
      if ( ! outputUnaryEvents( schema_name, udtr_sign, mother, sign_plus, *output_stream ) ) return false;
      if ( reference_stream && reference_prob_unary ) {
        double prob = 0.0;
        if ( ! unaryReference( schema_name, udtr_sign, mother, sign_plus, prob ) ) {
          return false;
        }
        reference_probability *= prob;
      }
      return outputParseTreeDtrs( udtr_tree );
    } else {
      // binary tree
      FSP mother = parse_tree.edgeSign();
      FSP sign_plus = parse_tree.signPlus();
      FSP schema_name = parse_tree.appliedSchema();
      ParseTree ldtr_tree = parse_tree.LTree();
      ParseTree rdtr_tree = parse_tree.RTree();
      FSP ldtr_sign = ldtr_tree.edgeSign();
      FSP rdtr_sign = rdtr_tree.edgeSign();
      if ( ! outputBinaryEvents( schema_name, ldtr_sign, rdtr_sign, mother, sign_plus, *output_stream ) ) return false;
      if ( reference_stream && reference_prob_binary ) {
        double prob = 0.0;
        if ( ! binaryReference( schema_name, ldtr_sign, rdtr_sign, mother, sign_plus, prob ) ) {
          return false;
        }
        reference_probability *= prob;
      }
      return outputParseTreeDtrs( ldtr_tree ) && outputParseTreeDtrs( rdtr_tree );
    }
  }

  bool ForestMaker::outputParseTreeRoot( ParseTree parse_tree ) {
    reference_probability = 1.0;
    if ( reference_stream && reference_prob_root ) {
      double prob = 0.0;
      if ( ! rootReference( parse_tree.edgeSign(), prob ) ) {
        return false;
      }
      reference_probability *= prob;
    }
    *output_stream << "1\t";
    if ( ! outputRootEvents( parse_tree.edgeSign(), *output_stream ) ) return false;
    if ( ! outputParseTreeDtrs( parse_tree ) ) return false;
    *output_stream << '\n';
    if ( reference_stream ) {
      *reference_stream << reference_probability << '\n';
    }
    return true;
  }

  bool ForestMaker::outputParseTree( ParseTree parse_tree, ostream& output_file, ostream& reference_file ) {
    output_stream = &output_file;
    reference_stream = &reference_file;
    return outputParseTreeRoot( parse_tree );
  }
  bool ForestMaker::outputParseTree( ParseTree parse_tree, ostream& output_file ) {
    output_stream = &output_file;
    reference_stream = NULL;
    return outputParseTreeRoot( parse_tree );
  }

  bool ForestMaker::outputForest( ostream& output_file, ostream& reference_file ) {
    output_stream = &output_file;
    reference_stream = &reference_file;
    return outputForestRoot();
  }
  bool ForestMaker::outputForest( ostream& output_file ) {
    output_stream = &output_file;
    reference_stream = NULL;
    return outputForestRoot();
  }

}

