/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file LexExtractor.cc
 *  @version Time-stamp: <2005-06-21 19:06:19 yusuke>
 *  Extracting lexical entries from derivations
 *
 **********************************************************************/

#include "LexExtractor.h"
#include "Initializer.h"
#include <liblilfes/utility.h>

#define LEX_EXTRACT_MODULE "mayz/lexextract"
#define LEXICAL_ENTRY_TEMPLATE "lexical_entry_template"
#define INVERSE_LEXICAL_RULE "reduce_lexical_template"
#define LEXEME_NAME "lexeme_name"

static lilfes::procedure* lexical_entry_template = NULL;
static lilfes::procedure* inverse_lexical_rule = NULL;
static lilfes::procedure* lexeme_name = NULL;

namespace mayz {

  using namespace lilfes;
  using namespace std;

  void LexExtractor::applyLexicalEntryTemplate( Derivation& terminal, lilfes::FSP lex_template ) {
    vector< FSP > args( 3 );
    args[ 0 ] = terminal.termWord().Copy();
    args[ 1 ] = terminal.derivSign().Copy();
    args[ 2 ] = lex_template;  // lex_template
    core_p cut_point = mach->GetCutPoint();
    if ( ! call_proc( *mach, lexical_entry_template, args ) ) {
      throw LexExtractorFailure( "lexical_entry_template/3 failed for " + terminal.termWord().input() + "/" + terminal.termWord().inputPos() );
    }
    mach->DoCut( cut_point );
    return;
  }
  void LexExtractor::applyInverseLexicalRule( Word word, lilfes::FSP lex_template, lilfes::FSP key, lilfes::FSP lexeme_template, lilfes::FSP lex_rule_list ) {
    vector< FSP > args( 5 );
    args[ 0 ] = word;
    args[ 1 ] = lex_template;
    args[ 2 ] = key;
    args[ 3 ] = lexeme_template;
    args[ 4 ] = lex_rule_list;
    core_p cut_point = mach->GetCutPoint();
    if ( ! call_proc( *mach, inverse_lexical_rule, args ) ) {
      throw LexExtractorFailure( INVERSE_LEXICAL_RULE "/5 failed for " + word.input() + "/" + word.inputPos() );
    }
    mach->DoCut( cut_point );
    return;
  }
  void LexExtractor::applyLexemeName( Word word, lilfes::FSP lexeme_template, std::string& name ) {
    vector< FSP > args( 4 );
    args[ 0 ] = word;
    args[ 1 ] = lexeme_template;
    args[ 2 ] = FSP( mach, num_templates++ );
    args[ 3 ] = FSP( mach );  // template name
    core_p cut_point = mach->GetCutPoint();
    if ( ! call_proc( *mach, lexeme_name, args ) ) {
      throw LexExtractorFailure( "lexeme_name/4 failed for " + word.input() + "/" + word.inputPos() );
    }
    mach->DoCut( cut_point );
    if ( ! lilfes_to_c< string >().convert( *mach, args[ 3 ], name ) ) {
      throw LexExtractorFailure( "4th argument of lexeme_name/4 must be string for " + word.input() + "/" + word.inputPos() );
    }
    return;
  }

  void LexExtractor::initialize() {
    if ( ! load_module( *mach, "lexicon extractor module", LEX_EXTRACT_MODULE ) ) {
      throw LexExtractorException( "Module \"" LEX_EXTRACT_MODULE "\" not found" );
    }
    if ( ! Initializer::initialize( mach ) ) {
      throw LexExtractorException( "initialization failed" );
    }
    lexical_entry_template = prepare_proc( module::UserModule(), LEXICAL_ENTRY_TEMPLATE, 3 );
    if ( ! lexical_entry_template ) {
      throw LexExtractorException( "Predicate " LEXICAL_ENTRY_TEMPLATE "/3 not found" );
    }
    inverse_lexical_rule = prepare_proc( module::UserModule(), INVERSE_LEXICAL_RULE, 5 );
    if ( ! inverse_lexical_rule ) {
      throw LexExtractorException( "Cannot find predicate: " INVERSE_LEXICAL_RULE "/5" );
    }
    lexeme_name = prepare_proc( module::UserModule(), LEXEME_NAME, 4 );
    if ( ! lexeme_name ) {
      throw LexExtractorException( "Cannot find predicate: " LEXEME_NAME "/4" );
    }
  }

  void LexExtractor::extractMain( int sentence_id, Derivation derivation ) {
    if ( derivation.isTerminal() ) {
      // terminal node
      FSP lex_template1( mach );
      applyLexicalEntryTemplate( derivation, lex_template1 );
      FSP lex_template( mach );
      if ( ! builtin::canonical_copy( *mach, lex_template1, lex_template ) ) {
        throw LexExtractorException( "canonical_copy/2 failed.  maybe lilfes's bug" );
      }
      FSP lexeme_template1( mach );
      FSP lex_rule_list( mach );
      FSP key( mach );
      applyInverseLexicalRule( derivation.termWord(), lex_template1, key, lexeme_template1, lex_rule_list );
      FSP lexeme_template( mach );
      if ( ! builtin::canonical_copy( *mach, lexeme_template1, lexeme_template ) ) {
        throw LexExtractorException( "canonical_copy/2 failed.  maybe lilfes's bug" );
      }
      LexTemplate lexeme_template_name( *mach );
      if ( ! temp_reverse_table.find( lexeme_template, lexeme_template_name ) ) {
        std::string name;
        applyLexemeName( derivation.termWord(), lexeme_template1, name );
        if ( ! ( lexeme_template_name.lexemeName().Unify( name )
                 && lexeme_template_name.lexicalRules().Coerce( nil ) ) ) {
          throw LexExtractorException( "making lex_template structure failed.  maybe lilfes's bug" );
        }
        if ( ( ! temp_reverse_table.insert( lexeme_template, lexeme_template_name ) )
             || ( ! temp_table->insert( lexeme_template_name, lexeme_template ) ) ) {
          throw LexExtractorException( "database update failed: template table" );
        }
      }
      if ( ! temp_table->incCount( lexeme_template_name ) ) {
        throw LexExtractorException( "database update failed: template count table" );
      }
      LexTemplate inflected_template_name( *mach );
      if ( ( ! inflected_template_name.lexemeName().Unify( lexeme_template_name.lexemeName() ) )
           || ( ! inflected_template_name.lexicalRules().Unify( lex_rule_list ) ) ) {
        throw LexExtractorException( "making lex_template structure.  maybe lilfes'f bug" );
      }
      //FSP word = derivation.termWord();
      //cerr << key.DisplayAVM() << template_name.DisplayAVM();
      if ( ! lexicon->add( key, lexeme_template_name ) ) {
        throw LexExtractorException( "database update failed: lexicon" );
      }
      if ( ! word_count->incCount( key ) ) {
        throw LexExtractorException( "database update failed: word count" );
      }
      //LexEntry lex_entry( template_name.Copy() );
      //if ( ! lex_entry.lexWord().Unify( word ) ) {
      //  return false;
      //}
      //cerr << lex_entry.DisplayAVM();
      if ( ( ! derivation.termTemplate().Unify( inflected_template_name ) )
           || ( ! derivation.lexentrySign().Unify( lex_template ) )
           || ( ! derivation.lexemeSign().Unify( lexeme_template ) ) ) {
        throw LexExtractorException( "derivation terminal structure failed.  maybe lilfes's bug" );
      }
      //lex_entry_list.push_back( lex_entry );
      lex_entry_list.push_back( derivation );
      return;
    } else {
      // internal node
      FSP dtrs = derivation.derivDtrs();
      while ( dtrs.GetType() == cons ) {
        extractMain( sentence_id, Derivation( dtrs.Follow( hd ) ) );
        dtrs = dtrs.Follow( tl );
      }
      return;
    }
  }

  void LexExtractor::extract( int sentence_id, Derivation derivation ) {
    ++num_all_derivations;
    lex_entry_list.resize( 0 );
    extractMain( sentence_id, derivation );
    ++num_success;
    num_total_words += lex_entry_list.size();
    if ( ! lexbank->insert( sentence_id, lex_entry_list ) ) {
      throw LexExtractorException( "database update failed: lexbank" );
    }
    return;
  }

  void LexExtractor::showStatistics( ostream& os, const std::string& indent ) const {
    os << indent << "# templates = " << num_templates << endl;
    os << indent << "# total words = " << num_total_words << endl;
  }

}

