/*
 * $Id: Parser.cc,v 1.4 2005/07/04 14:30:00 yusuke Exp $
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 */

#include "Parser.h"
#include "Grammar.h"
#include "ParserBuiltin.h"
#include <liblilfes/profile.h>
#include <liblilfes/utility.h>
#include <iostream>


namespace up {
    
    using namespace lilfes;

    void ParserBasic::startProf(const char* s) {
        lilfes::start_prof(s);
    }
    
    int ParserBasic::stopProf(const char* s) {
        return lilfes::stop_prof(s);
    }

    bool ParserBasic::init(machine *m, Grammar*) {
	mach = m;
        Builtin::setParser(this);
        setParserName(std::string("UP"));
        setParserVersion(std::string("0.8"));
        return true;
    }

    bool ParserBasic::parseSetup(WordLattice& wordlattice) {
        P2("ParserBasic::parseSetup");
        setWordLattice(wordlattice);
        WordLattice::iterator it = wordlattice.begin(), last = wordlattice.end();
        int sentlen = 0;
        for(; it != last ; it++) {
            int rp = (*it)->getRightPosition();
            sentlen = rp > sentlen ? rp : sentlen;
        }
        setSentenceLength(sentlen);
        if(getSentenceLength() > getLimitWords()) return false;
        clear();
        return true;
    }
    
    bool ParserBasic::parsePreproc(std::string& __sentence, WordLattice& wordlattice) {
        P2("ParserBasic::parsePreproc");
        setSentence(__sentence);
        IPTrailStack iptrail(mach);
	FSP wl(mach);
	FSP s(mach, getSentence().c_str());
	bool loop = getGrammar()->sentence2WordLattice(s, wl);
	if(! loop) return false;
        while( wl.GetType()->IsSubType(cons) ) {
            wordlattice.push_back(getFspMaker()->newExtent(wl.Follow(hd)));
            wl = wl.Follow(tl);
        }
        return true;
    }
    
//     bool ParserBasic::parseWithBracket(std::string __sentence) {
//         IPTrailStack iptrail(mach);
        
//         WordLattice wordlattice;
        
//             // preprocessing
//         static const char *str_preprocessing = "preprocessing"; startProf(str_preprocessing);
//         bool preprocessing_succ = (parsePreproc(__sentence, wordlattice) &&
//                                    parseSetup(wordlattice));
//         getChart()->makeDead(wordlattice, getSentenceLength());
//         setPreprocessingTime(stopProf(str_preprocessing));
//         if(! preprocessing_succ ) return false;

//             // lexicon lookup
//         static const char* str_lexicon_lookup = "lexicon_lookup"; startProf(str_lexicon_lookup);
//         bool lexicon_lookup_succ = parseLex(wordlattice);
//         setLexiconLookupTime(stopProf(str_lexicon_lookup));
//         if(! lexicon_lookup_succ) return false;
        
//             // parsing
//         static const char* str_parsing = "parsing"; startProf(str_parsing);
//         bool parsing_succ = parsePhrase();
//         setParsingTime(stopProf(str_parsing));
//         if(! parsing_succ) return false;
        
//         return true;
//     }

    bool ParserBasic::parse(std::string __sentence) {
        P2("ParserBasic::parse");
        IPTrailStack iptrail(mach);

        WordLattice wordlattice;
        
            // preprocessing
        static const char* str_preprocessing = "preprocessing"; startProf(str_preprocessing);
        bool preprocessing_succ = parsePreproc(__sentence, wordlattice);
        if ( ! preprocessing_succ ) {
          setParseStatus( PREPROC_ERROR );
        } else {
          preprocessing_succ = parseSetup(wordlattice);
          if ( ! preprocessing_succ ) {
            setParseStatus( TOO_LONG );
          } else {
            getChart()->makeDead(wordlattice, getSentenceLength());
            getChart()->inhibitLargeConstituents(getLimitConstituentSize(), getSentenceLength());
          }
        }
        setPreprocessingTime(stopProf(str_preprocessing));
        if(! preprocessing_succ ) return false;

            // lexicon lookup
        static const char* str_lexicon_lookup = "lexicon_lookup"; startProf(str_lexicon_lookup);
        bool lexicon_lookup_succ = parseLex(wordlattice);
        setLexiconLookupTime(stopProf(str_lexicon_lookup));
        if(! lexicon_lookup_succ) {
          setParseStatus( EDGE_LIMIT );
          return false;
        }
        
            // parsing
        static const char* str_parsing = "parsing"; startProf(str_parsing);
        bool parsing_succ = parsePhrase();
        setParsingTime(stopProf(str_parsing));
        if(! parsing_succ) {
          setParseStatus( EDGE_LIMIT );
          return false;
        }

        setParseStatus( SUCCESS );
        return true;
    }
    
    bool ParserBasic::parseWordLattice(WordLattice wordlattice) {
        P2("ParserBasic::parseWordLattice");
        IPTrailStack iptrail(mach);
        
            // preprocessing
        static const char* str_preprocessing = "preprocessing"; startProf(str_preprocessing);
        bool preprocessing_succ = parseSetup(wordlattice);
        getChart()->makeDead(wordlattice, getSentenceLength());
        getChart()->inhibitLargeConstituents(getLimitConstituentSize(), getSentenceLength());
        setPreprocessingTime(stopProf(str_preprocessing));
        if(! preprocessing_succ ) {
          setParseStatus( TOO_LONG );
          return false;
        }

            // lexicon lookup
        static const char* str_lexicon_lookup = "lexicon_lookup"; startProf(str_lexicon_lookup);
        bool lexicon_lookup_succ = parseLex(wordlattice);
        setLexiconLookupTime(stopProf(str_lexicon_lookup));
        if(! lexicon_lookup_succ) {
          setParseStatus( EDGE_LIMIT );
          return false;
        }
        
            // parsing
        static const char* str_parsing = "parsing"; startProf(str_parsing);
        bool parsing_succ = parsePhrase();
        setParsingTime(stopProf(str_parsing));
        if(! parsing_succ) {
          setParseStatus( EDGE_LIMIT );
          return false;
        }

        setParseStatus( SUCCESS );
        return true;
    }
    
} // end of namespace up

/*
 * $Log: Parser.cc,v $
 * Revision 1.4  2005/07/04 14:30:00  yusuke
 *
 * fixed segmentation fault when brackets are specified.
 *
 * Revision 1.3  2005/06/21 10:17:12  yusuke
 *
 * copyright 񤭤ޤ
 *
 * Revision 1.2  2005/06/16 07:25:36  yusuke
 *
 * integrate-up 硥
 * version 2.0
 *
 * Revision 1.1.2.11  2005/06/11 15:10:50  yusuke
 *
 * constituent 礭¤Ǥ褦ˤ
 *
 * Revision 1.1.2.10  2005/06/08 05:59:42  ninomi
 * P2ץեĤʤޤ
 *
 * Revision 1.1.2.9  2005/06/08 03:30:43  yusuke
 *
 * parse_with_brackets ѻߡ
 * get_parse_status 
 *
 * Revision 1.1.2.8  2005/06/07 08:49:17  ninomi
 * Parser::init(machine*)init(machine*, Grammar*)Ȥޤ
 *
 * Revision 1.1.2.7  2005/06/07 07:37:28  ninomi
 * Parserݥ饹ľޤ
 *
 * Revision 1.1.2.6  2005/06/06 12:01:02  yusuke
 *
 * Ȥꤢ sample_hpsg  warning Фʤ褦ˤޤ
 *
 * Revision 1.1.2.5  2005/06/06 09:44:28  ninomi
 * builtinޤդ䤷ޤ
 *
 * Revision 1.1.2.4  2005/06/06 08:53:41  ninomi
 * IPTrailStackطľޤlilfespredicate callʤޤ
 * quick checkidҤˤΤnameǤҤ褦ˤޤ
 *
 * Revision 1.1.2.3  2005/06/06 03:37:37  ninomi
 * gotoʸޤ
 *
 * Revision 1.1.2.2  2005/06/05 14:10:48  ninomi
 * ѥ뤬Ȥ褦ˤʤޤ
 *
 * Revision 1.1.2.1  2005/06/05 10:12:18  ninomi
 * upmayz/srcβ˰ܤϤޤ
 *
 * Revision 1.13  2004/11/16 09:14:07  ninomi
 * Artistic LicenseƬˤĤޤ
 *
 * Revision 1.12  2004/07/02 09:46:24  ninomi
 * start_profstop_profll˰ܤޤ
 *
 * Revision 1.11  2004/06/30 13:38:23  ninomi
 * ʸĹsetLimitWordsǡ票åsetLimitEdgesǥѡǤʤ
 * ФĤǤѹǤ褦ˤޤ
 *
 * Revision 1.10  2004/06/30 07:42:13  ninomi
 * up 0.2˰ܹԤ뤿0.1θŤե
 *
 * Revision 1.9  2004/06/14 14:06:36  yusuke
 *
 * stopProf ñ̤ǤפʤäΤ
 *
 * Revision 1.8  2004/05/20 23:21:00  ninomi
 * cup, pcup, qcup, pqcupСǤcup, qcupưƤޤ
 *
 * Revision 1.7  2004/05/18 17:05:00  ninomi
 * profileʸХʤ(string->static const char*)
 *
 * Revision 1.6  2004/05/18 16:40:23  ninomi
 * PCKYäޤ
 *
 * Revision 1.5  2004/05/17 12:20:59  ninomi
 * QuickCheckǽܤޤ
 *
 * Revision 1.4  2004/05/17 08:21:16  ninomi
 * GrammarIFϰ̤Grammar˴ؤ뵡ǽ󶡤뤳ȤˤơѾطʤ
 * ޤ
 *
 * GrammarIFBasic -> GrammarIF
 * GrammarIFQC -> QuickCheckIF
 *
 * Revision 1.3  2004/05/15 18:59:04  ninomi
 * 󥯤ǤޤǤǤޤ
 *
 * Revision 1.2  2004/05/15 17:49:55  ninomi
 * *** empty log message ***
 *
 * Revision 1.1  2004/05/15 16:16:44  ninomi
 * ParserCKYΤɤΥѡˤⶦ̤Ȼפ륳ݡͥȤParserBasic
 * 褦ˤ
 *
 */
