/*
 * $Id: Parser.h,v 1.3 2005/06/21 10:17:12 yusuke Exp $
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 */

#ifndef __Parser_h
#define __Parser_h

#include "Chart.h"
#include "Fs.h"
#include "Fsp.h"
#include "Edge.h"
#include "Grammar.h"
#include "Extent.h"
#include <vector>
#include <string>

namespace up {
    
    class Parser {
    public:
      enum ParseStatus {
        NOT_PARSED_YET, SUCCESS, PREPROC_ERROR,
        TOO_LONG, EDGE_LIMIT, UNKNOWN
      };
    public:
//// constructor and destructor        
        Parser() {}
        virtual ~Parser() {}
        
//// primitive methods        
        virtual int getSentenceLength() = 0;
        
        virtual int getLimitEdges() = 0;
        virtual void setLimitEdges(int x) = 0;
        
        virtual int getLimitWords() = 0;
        virtual void setLimitWords(int x) = 0;

        virtual ParseStatus getParseStatus() = 0;
        
        virtual const std::string& getSentence() = 0;
        
        virtual const WordLattice& getWordLattice() = 0;

        virtual int getPreprocessingTime() = 0;
        virtual int getLexiconLookupTime() = 0;
        virtual int getParsingTime() = 0;

        virtual std::string& getGrammarName() = 0;
        virtual void setGrammarName(std::string gn) = 0;
        
        virtual std::string& getGrammarVersion() = 0;
        virtual void setGrammarVersion(std::string gv) = 0;

        virtual std::string& getParserName() = 0;
        virtual std::string& getParserVersion() = 0;

//// retrieve and set components
        virtual void setGrammar(Grammar* g) = 0;
        virtual Grammar* getGrammar() = 0;
        
        virtual void setChart(Chart* c) = 0;
        virtual Chart* getChart() = 0;
        
        virtual FspMaker* getFspMaker() = 0;

//// parser methods
        virtual Edge* getEdge(eserial es) = 0;
        virtual Link* getLink(lserial ls) = 0;
        virtual int getEdgeNumber() = 0;
        virtual std::vector<Edge*>& getEdges(int, int) = 0;
        
        virtual bool init(lilfes::machine*, Grammar*) = 0;
        virtual bool term() = 0;
        virtual void clear() = 0;
        
        virtual bool parse(std::string sentence) = 0;
//         virtual bool parseWithBracket(std::string sentence) = 0;
        virtual bool parseWordLattice(WordLattice wordlattice) = 0;
    };
    
    class ParserBasic : public Parser {
    private:
        int limit_edges;
        int limit_words;
        int sentence_length;

        ParseStatus parse_status;

        std::string sentence;
        WordLattice word_lattice;

        int preprocessing_time;
        int lexicon_lookup_time;
        int parsing_time;

        std::string grammar_name;
        std::string grammar_version;
        
        std::string parser_name;
        std::string parser_version;

        int limit_constituent_size;
        
    protected:
        virtual void setSentenceLength(int x) { sentence_length = x; }
        virtual void setSentence(const std::string& s) { sentence = s; }
        virtual void setWordLattice(const WordLattice& w) { word_lattice = w; }
        
        ParseStatus setParseStatus( ParseStatus s ) {
          ParseStatus prev = parse_status;
          parse_status = s;
          return prev;
        }
        
        virtual void setPreprocessingTime(int x) { preprocessing_time = x; }
        virtual void setLexiconLookupTime(int x) { lexicon_lookup_time = x; }
        virtual void setParsingTime(int x) { parsing_time = x; }

        virtual void setParserName(std::string pn) { parser_name = pn; }
        virtual void setParserVersion(std::string pv) { parser_version = pv; }
        
	lilfes::machine *mach;
	lilfes::machine* getMachine() { return mach; }
        void setMachine(lilfes::machine* m) { mach = m; }

        void startProf(const char* s);
        int stopProf(const char* s);

        virtual bool parsePreproc(std::string& __sentence, WordLattice& wordlattice);
        virtual bool parseSetup(WordLattice& wordlattice);
        virtual bool parseLex(const WordLattice& wordlattice) = 0;
        virtual bool parsePhrase() = 0;
        
    public:
        ParserBasic() : limit_edges(0), limit_words(0), sentence_length(0), parse_status( NOT_PARSED_YET ), limit_constituent_size(0) {}
        ~ParserBasic() {}
        
        int getSentenceLength() {return sentence_length; }
        
        int getLimitEdges() {return limit_edges; }
        void setLimitEdges(int x) { limit_edges = x; }
        
        int getLimitWords() {return limit_words; }
        void setLimitWords(int x) { limit_words = x; }

        virtual void setLimitConstituentSize(int x) { limit_constituent_size = x; }
        virtual int getLimitConstituentSize() { return limit_constituent_size; }

        ParseStatus getParseStatus() { return parse_status; }
        
        const std::string& getSentence() { return sentence; }
        const WordLattice& getWordLattice() { return word_lattice; }

        int getPreprocessingTime() { return preprocessing_time; }
        int getLexiconLookupTime() { return lexicon_lookup_time; }
        int getParsingTime() { return parsing_time; }

        std::string& getGrammarName() { return grammar_name; }
        void setGrammarName(std::string gn) { grammar_name = gn; }
        
        std::string& getGrammarVersion() { return grammar_version; }
        void setGrammarVersion(std::string gv) { grammar_version = gv; }

        std::string& getParserName() { return parser_name; }
        std::string& getParserVersion() { return parser_version; }
        
        void setGrammar(Grammar*) {}
        Grammar* getGrammar() {return 0; }
        
        void setChart(Chart* ) {}
        Chart* getChart() {return 0; }
        
        FspMaker* getFspMaker() { return 0; }
        
        bool init(lilfes::machine*, Grammar*);
        bool term() { return true; }
        bool parse(std::string __sentence);
//         bool parseWithBracket(std::string __sentence);
        bool parseWordLattice(WordLattice wordlattice);

        friend class Builtin;
    };
    
} // end of namespace up
#endif // __Parser_h

/*
 * $Log: Parser.h,v $
 * Revision 1.3  2005/06/21 10:17:12  yusuke
 *
 * copyright 񤭤ޤ
 *
 * Revision 1.2  2005/06/16 07:25:36  yusuke
 *
 * integrate-up 硥
 * version 2.0
 *
 * Revision 1.1.2.9  2005/06/11 15:10:50  yusuke
 *
 * constituent 礭¤Ǥ褦ˤ
 *
 * Revision 1.1.2.8  2005/06/08 03:30:43  yusuke
 *
 * parse_with_brackets ѻߡ
 * get_parse_status 
 *
 * Revision 1.1.2.7  2005/06/07 08:49:17  ninomi
 * Parser::init(machine*)init(machine*, Grammar*)Ȥޤ
 *
 * Revision 1.1.2.6  2005/06/07 07:37:28  ninomi
 * Parserݥ饹ľޤ
 *
 * Revision 1.1.2.5  2005/06/06 12:40:50  yusuke
 *
 * ȤꤢѡǤ褦ˤʤä
 *
 * Revision 1.1.2.4  2005/06/06 09:44:28  ninomi
 * builtinޤդ䤷ޤ
 *
 * Revision 1.1.2.3  2005/06/06 08:53:41  ninomi
 * IPTrailStackطľޤlilfespredicate callʤޤ
 * quick checkidҤˤΤnameǤҤ褦ˤޤ
 *
 * Revision 1.1.2.2  2005/06/05 14:10:48  ninomi
 * ѥ뤬Ȥ褦ˤʤޤ
 *
 * Revision 1.1.2.1  2005/06/05 10:12:19  ninomi
 * upmayz/srcβ˰ܤϤޤ
 *
 * Revision 1.16  2004/11/16 09:13:54  ninomi
 * Artistic LicenseƬˤĤޤ
 *
 * Revision 1.15  2004/07/09 02:11:54  yusuke
 *
 * Parser::getWordLattice 
 * iterative parsing ǡbeam ѤƤʤȤ
 * parseLexiconLookup ʤ褦ˤ
 *
 * Revision 1.14  2004/07/06 13:05:50  ninomi
 * Parser::term()ɲäޤlilfesterminator::PerformAll()¹
 * ɬ¹ԤʤȤʤ
 *
 * Revision 1.13  2004/06/30 13:38:15  ninomi
 * ʸĹsetLimitWordsǡ票åsetLimitEdgesǥѡǤʤ
 * ФĤǤѹǤ褦ˤޤ
 *
 * Revision 1.12  2004/06/30 07:41:48  ninomi
 * up 0.2˰ܹԤ뤿0.1θŤե
 *
 * Revision 1.11  2004/05/20 23:24:58  ninomi
 * СǤcup, qcupưƤޤpcup, pqcupϥХȤ档
 *
 * Revision 1.10  2004/05/17 12:20:53  ninomi
 * QuickCheckǽܤޤ
 *
 * Revision 1.9  2004/05/15 17:49:48  ninomi
 * *** empty log message ***
 *
 * Revision 1.8  2004/05/15 16:16:38  ninomi
 * ParserCKYΤɤΥѡˤⶦ̤Ȼפ륳ݡͥȤParserBasic
 * 褦ˤ
 *
 * Revision 1.7  2004/05/14 12:41:30  ninomi
 * ѥ뤬Ȥ褦ˤʤޤ
 *
 * Revision 1.6  2004/05/14 05:01:30  ninomi
 * GrammarIFNaiveȤ̾ΤGrammarIFBasicȤ̾Τˤޤ
 *
 * Revision 1.5  2004/05/14 04:33:29  ninomi
 * Ǥ
 *
 * Revision 1.4  2004/05/13 12:49:02  yusuke
 *
 * Parser.h  GrammarIF ȤäƤΤ include 褦ˤ
 * ˤȤʤGrammarIF ѥ̤褦ˤ
 *
 * Revision 1.3  2004/05/13 12:45:01  yusuke
 *
 * Parser.h  Edge 饹ȤäƤΤǡEdge.h  include 
 * ˤȤʤEdge 饹ѥ̤褦ˤ
 *
 * Revision 1.2  2004/05/13 07:37:06  ninomi
 * ä
 *
 * Revision 1.1  2004/05/13 07:07:23  ninomi
 * ʥեɲǤ
 *
 */

