/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file forestmaker_main.cc
 *  @version Time-stamp: <2005-06-23 17:12:58 yusuke>
 *  Main program of forestmaker
 *
 **********************************************************************/

#include "ForestMaker.h"
#include "Derivbank.h"
#include <liblilfes/builtin.h>
#include "ParserCky.h"
#include <string>
#include <set>

using namespace std;
using namespace lilfes;
using namespace up;
using namespace mayz;

enum file_type {
  RAW, BZ2, GZ
};

static string model_name;
static string grammar_name;
static string derivbank_name;
static string uevent_file_name;
static string reference_file_name;
static int limit_sentences = 0;
static int verbose_level = 0;

static set< int > target_sentences;

//////////////////////////////////////////////////////////////////////

void help_message() {
  *error_stream << "Usage: forestmaker [options] model_name grammar_module derivbank_file uevent_file" << endl;
  *error_stream << "  model_name: name of the model to be made" << endl;
  *error_stream << "  grammar_module: LiLFeS module in which the grammar and event extraction predicates are defined" << endl;
  *error_stream << "  derivbank_file: derivation database file" << endl;
  *error_stream << "  uevent_file: output file" << endl;
  *error_stream << "Options:" << endl;
  *error_stream << "  -r filename:\treference distribution file" << endl;
  *error_stream << "  -n num:\tlimit number of sentences (default: " << limit_sentences << ")" << endl;
  *error_stream << "  -v:\t\tprint verbose messages" << endl;
  *error_stream << "  -vv:\t\tprint many verbose messages" << endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  char** end = argv + argc;
  ++argv;
  for ( ; argv != end; ++argv ) {
    if ( (*argv)[ 0 ] == '-' ) {
      // option without argument
      if ( std::strcmp( (*argv), "-v" ) == 0 ) {
        verbose_level = 1;
      } else if ( std::strcmp( (*argv), "-vv" ) == 0 ) {
        verbose_level = 2;
      } else {
        // option with argument
        if ( argv + 1 == end ) {
          *error_stream << *argv << " option requires additional argument" << endl;
          return false;
        }
        if ( std::strcmp( (*argv), "-r" ) == 0 ) {
          reference_file_name = *(++argv);
        } else if ( std::strcmp( (*argv), "-n" ) == 0 ) {
          limit_sentences = strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "--target-sentence" ) == 0 ) {
          target_sentences.insert( strtol( *(++argv), NULL, 0 ) );
        } else {
          *error_stream << "Unknown option: " << *argv << endl;
          return false;
        }
      }
    } else {
      // arguments
      if ( end - argv != 4 ) {
        *error_stream << "forestmaker requires four arguments" << endl;
        return false;
      }
      model_name = argv[ 0 ];
      grammar_name = argv[ 1 ];
      derivbank_name = argv[ 2 ];
      uevent_file_name = argv[ 3 ];
      return true;
    }
  }
  *error_stream << "forestmaker requires four arguments" << endl;
  return false;
}

//////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{

  *error_stream << "forestmaker (MAYZ " << MAYZ_VERSION << ")" << endl;
  *error_stream << MAYZ_COPYRIGHT << endl;

  try {
    // Initialize everything
    Initializer::PerformAll();

    // show help message
    if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
      help_message();
      return 0;
    }

    // command-line arguments
    if ( ! analyze_arguments( argc, argv ) ) {
      help_message();
      return 1;
    }

    // Initialize a machine object
    machine mach;

    // file type prediction
    file_type uevent_file_type = RAW;
    if ( uevent_file_name.substr( uevent_file_name.size() - 4, std::string::npos ) == ".bz2" ) {
      uevent_file_type = BZ2;
    } else if ( uevent_file_name.substr( uevent_file_name.size() - 3, std::string::npos ) == ".gz" ) {
      uevent_file_type = GZ;
    }

    // Load grammar module
    *error_stream << "Loading grammar module \"" << grammar_name << "\"... ";
    if ( ! load_module( mach, "grammar module", grammar_name ) ) {
      *error_stream << "grammar module not found" << endl;
      return 1;
    }
    if ( ! load_module( mach, "parser module", "mayz/up" ) ) {
      *error_stream << "parser module \"mayz/up\" not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    // Load derivbank
    *error_stream << "Loading derivbank \"" << derivbank_name << "\"... ";
    Derivbank derivbank( &mach );
    if ( ! derivbank.load( derivbank_name ) ) {
      *error_stream << "derivbank not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    // Initialize parser
    *error_stream << "Initializing forest maker...\n";
    FMGrammar fm_grammar;
    if ( ! fm_grammar.init( &mach ) ) {
      *error_stream << "initialization of a grammar failed." << endl;
      return 1;
    }
    ForestMaker forest_maker( model_name );
//     forest_maker.setBeamThresholdNum( 0 );
//     forest_maker.setBeamThresholdWidth( 0.0 );
    if( ! forest_maker.init( &mach, &fm_grammar ) ) {
      *error_stream << "failed." << endl;
      return 1;
    }
    forest_maker.enableQuickCheck();
    *error_stream << "done." << endl;
    *error_stream << "Grammar: " << forest_maker.getGrammarName() << " - " << forest_maker.getGrammarVersion() << endl;
    *error_stream << "Limit of sentence length: " << forest_maker.getLimitWords() << endl;
    *error_stream << "Limit of edge number: " << forest_maker.getLimitEdges() << endl;

    // Start lilfes command
    *error_stream << "Start making feature forests" << endl;
    {
      std::auto_ptr< std::ostream > unfiltered_event_file( NULL );
      if ( uevent_file_type == RAW ) {
        unfiltered_event_file.reset( new std::ofstream( uevent_file_name.c_str() ) );
      } else if ( uevent_file_type == BZ2 ) {
        unfiltered_event_file.reset( new obfstream( uevent_file_name.c_str() ) );
      } else if ( uevent_file_type == GZ ) {
        unfiltered_event_file.reset( new ogfstream( uevent_file_name.c_str() ) );
      } else {
        *error_stream << "Unknown compression type: " << uevent_file_name << endl;
        return 1;
      }
      if ( ! (*unfiltered_event_file) ) {
        *error_stream << "Cannot open unfiltered event file: " << uevent_file_name << endl;
        return 1;
      }
      std::auto_ptr< std::ostream > reference_file( NULL );
      if ( ! reference_file_name.empty() ) {
        reference_file.reset( new std::ofstream( reference_file_name.c_str() ) );
      }
      //for ( size_t deriv_id = 1; deriv_id <= derivbank.maxID(); ++deriv_id ) {
      for ( Derivbank::iterator deriv_it = derivbank.begin();
            deriv_it != derivbank.end();
            ++deriv_it ) {
        core_p trailp = mach.SetTrailPoint( NULL );
        int deriv_id = deriv_it->first;
        if ( limit_sentences > 0 && deriv_id > limit_sentences ) break;
        if ( ! target_sentences.empty() && target_sentences.find( deriv_id ) == target_sentences.end() ) continue;
        Derivation derivation = deriv_it->second;
        WordLattice wordlattice;
        ParseTree parse_tree( mach );
        *error_stream << "No. " << deriv_id << ": ";
        if ( ! forest_maker.makeWordLattice( derivation, wordlattice ) ) {
          *error_stream << "fm_derivation_to_word_lattice/2 failed" << endl;
        } else if ( ! forest_maker.derivationToParseTree( derivation, parse_tree ) ) {
          *error_stream << "derivation_to_parse_tree/2 failed" << endl;
        } else if ( ! forest_maker.parseWordLattice( wordlattice ) ) {
          *error_stream << "parse error: ";
          switch( forest_maker.getParseStatus() ) {
          case Parser::SUCCESS:
            *error_stream << "no successful parse" << endl;
            break;
          case Parser::PREPROC_ERROR:
            *error_stream << "preprocessing error" << endl;
            break;
          case Parser::TOO_LONG:
            *error_stream << "too long sentence (" << forest_maker.getSentenceLength() << " words)" << endl;
            break;
          case Parser::EDGE_LIMIT:
            *error_stream << "edge limit exceeded" << endl;
            break;
          default:
            *error_stream << "unknown reason" << endl;
          }
        } else if ( forest_maker.getEdges( 0, forest_maker.getSentenceLength() ).empty() ) {
          *error_stream << "no successful parse" << endl;
        } else {
          *unfiltered_event_file << "event_" << deriv_id << '\n';
          if ( reference_file.get() != NULL ) {
            // with reference distribution
            *reference_file << "event_" << deriv_id << '\n';
            if ( ! forest_maker.outputParseTree( parse_tree, *unfiltered_event_file, *reference_file ) ) {
              *error_stream << "output parse tree failed" << endl;
              return 1;
            }
            if ( ! forest_maker.outputForest( *unfiltered_event_file, *reference_file ) ) {
              *error_stream << "output feature forest failed" << endl;
              return 1;
            }
            *reference_file << '\n';
          } else {
            // without reference distribution
            if ( ! forest_maker.outputParseTree( parse_tree, *unfiltered_event_file ) ) {
              *error_stream << "output parse tree failed" << endl;
              return 1;
            }
            if ( ! forest_maker.outputForest( *unfiltered_event_file ) ) {
              *error_stream << "output feature forest failed" << endl;
              return 1;
            }
          }
          *unfiltered_event_file << '\n';
          *error_stream << "success" << endl;
        }
        mach.TrailBack( trailp );
        //*error_stream << "a4" << endl;
      }
    }
    *error_stream << "done." << endl;

    // Termination process
    if( ! forest_maker.term() ) {
        *error_stream << "parser failed to terminate" << endl;
        return 1;
    }
    if( ! fm_grammar.term() ) {
      *error_stream << "termination of the grammar failed" << endl;
      return 1;
    }
    Terminator::PerformAll();

    return 0;
  } 
  catch (bad_alloc) {
    cerr << "Memory allocation failed, forestmaker terminated." << endl;
    throw;
  }
  catch (exception &e) {
    cerr << "Fatal error: " << e.what() << endl;
    throw;
  }
}

