/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file lex2html_main.cc
 *  @version Time-stamp: <2005-06-21 19:23:54 yusuke>
 *  Main program of lex2html
 *
 **********************************************************************/

#include "Lexbank.h"
#include "Lexicon.h"
#include "TemplateTable.h"
#include "Word.h"
#include "Derivation.h"
#include <liblilfes/utility.h>
#include <liblilfes/xml.h>
#include <iostream>
#include <algorithm>
#include <functional>

using namespace std;
using namespace lilfes;
using namespace mayz;

string grammar_module;
string lexicon_name;
string temptable_name;
string lexbank_name;
string output_dir( "." );
string output_prefix;
string stylesheet;
int verbose_level = 0;

procedure* lexicon_lookup_key = NULL;
procedure* unknown_word_lookup_key = NULL;
procedure* lex_template_label = NULL;
procedure* sign_label = NULL;
procedure* lexicon_key_label = NULL;
procedure* word_label = NULL;

int lex_file_size = 100;
int temp_file_size = 20;
int lexeme_file_size = 100;
int sentence_file_size = 1000;

string charset_string;

//////////////////////////////////////////////////////////////////////

class lex2html_exception : public exception {
private:
  string message;
public:
  lex2html_exception( const string& m ) : message( m ) {}
  virtual ~lex2html_exception() throw () {}
  const char* what() const throw () { return message.c_str(); }
};

//////////////////////////////////////////////////////////////////////

void inc_count_db( FSArray< FSP, int >& db, FSP key ) {
  int count = 0;
  db.find( key, count );
  if ( ! db.insert( key, ++count ) )
    throw lex2html_exception( "database access error (inc_count_db)" );
  return;
}

void update_occurrence_db( FSArray< FSP, pair< int, FSP > >& db, FSP key, int id, FSP val ) {
  pair< int, FSP > dummy;
  if ( ! db.find( key, dummy ) ) {
    if ( ! db.insert( key, make_pair( id, val ) ) )
      throw lex2html_exception( "database access error (update_occurrence_db)" );
  }
}

void update_expand_db( FSArray< FSP, vector< FSP > >& db, FSP key, FSP val ) {
  vector< FSP > expand_list;
  db.find( key, expand_list );
  bool flag = false;
  for ( vector< FSP >::iterator it = expand_list.begin();
        it != expand_list.end();
        ++it ) {
    if ( builtin::equivalent( *(db.getMachine()), *it, val ) ) {
      flag = true;
      break;
    }
  }
  if ( ! flag ) {
    expand_list.push_back( val );
    if ( ! db.insert( key, expand_list ) ) {
      throw lex2html_exception( "database access error (update_expand_db)" );
    }
  }
}

//////////////////////////////////////////////////////////////////////

FSP get_lexicon_key( machine* mach, FSP word ) {
  vector< FSP > args( 2 );
  args[ 0 ] = word;
  args[ 1 ] = FSP( mach );
  if ( ! call_proc( *mach, lexicon_lookup_key, args ) )
    throw lex2html_exception( "calling lexicon_lookup_key/2 failed" );
  return args[ 1 ];
}

FSP get_unknown_word_key( machine* mach, FSP word ) {
  vector< FSP > args( 2 );
  args[ 0 ] = word;
  args[ 1 ] = FSP( mach );
  if ( ! call_proc( *mach, unknown_word_lookup_key, args ) )
    throw lex2html_exception( "calling unknown_word_lookup_key/2 failed" );
  return args[ 1 ];
}

string get_template_name( machine* mach, LexTemplate temp ) {
  string name;
  vector< FSP > args( 2 );
  args[ 0 ] = temp;
  args[ 1 ] = FSP( *mach );
  if ( call_proc( *mach, lex_template_label, args ) &&
       args[ 1 ].IsString() ) {
    name = args[ 1 ].ReadString();
  }
  return name;
}

string get_lexeme_name( machine*, FSP lexeme ) {
  string name;
  if ( lexeme.IsString() ) {
    name = lexeme.ReadString();
  }
  return name;
}

string get_sign_symbol( machine* mach, FSP sign ) {
  string name;
  vector< FSP > args( 2 );
  args[ 0 ] = sign;
  args[ 1 ] = FSP( *mach );
  if ( call_proc( *mach, sign_label, args ) &&
       args[ 1 ].IsString() ) {
    name = args[ 1 ].ReadString();
  }
  return name;
}

string get_lexkey_symbol( machine* mach, FSP word ) {
  string name;
  vector< FSP > args( 2 );
  args[ 0 ] = word;
  args[ 1 ] = FSP( *mach );
  if ( call_proc( *mach, lexicon_key_label, args ) &&
       args[ 1 ].IsString() ) {
    name = args[ 1 ].ReadString();
  }
  return name;
}

string get_word_symbol( machine* mach, FSP word ) {
  string name;
  vector< FSP > args( 2 );
  args[ 0 ] = word;
  args[ 1 ] = FSP( *mach );
  if ( call_proc( *mach, word_label, args ) &&
       args[ 1 ].IsString() ) {
    name = args[ 1 ].ReadString();
  }
  return name;
}

int get_word_id( machine* mach, FSArray< FSP, int >& word_id_map, Word word ) {
  FSP key = get_lexicon_key( mach, word );
  int word_id = 0;
  if ( word_id_map.find( key, word_id ) ) return word_id;
  key = get_unknown_word_key( mach, word );
  if ( word_id_map.find( key, word_id ) ) return word_id;
  return -1;
}

string get_file_name( const string& name, int id ) {
  ostringstream str;
  str << output_prefix << "-" << name << "-" << id << ".html";
  return str.str();
}

string get_file_name( const string& name ) {
  ostringstream str;
  str << output_prefix << "-" << name << ".html";
  return str.str();
}

string get_output_file_name( const string& name, int id ) {
  return output_dir + "/" + get_file_name( name, id );
}

string get_output_file_name( const string& name ) {
  return output_dir + "/" + get_file_name( name );
}

void output_index_link( ostream* os ) {
  *os << "<hr>\n<a href=\"" << output_prefix << "-index.html\">Index</a>\n<hr>\n";
}

//////////////////////////////////////////////////////////////////////

class HTMLFile : public ofstream {
private:
  string filename;
  string title;
  string stylesheet;
  string charset_string;
protected:
  void outputHeader() {
    *this << "<!doctype html public \"-//w3c//dtd html 4.0//en\">\n";
    *this << "<html>\n";
    *this << "<head>\n";
    *this << "<meta http-equiv=\"Content-Type\" content=\"text/html";
    if ( ! charset_string.empty() ) {
      *this << "; charset=" << charset_string;
    }
    *this << "\">\n";
    if ( ! stylesheet.empty() ) {
      ifstream css( stylesheet.c_str() );
      if ( css ) {
        *this << "<style type=\"text/css\">\n<!--\n";
        while ( true ) {
          string css_line;
          getline( css, css_line );
          if ( ! css ) break;
          *this << css_line << '\n';
        }
        *this << "-->\n</style>\n";
      }
    }
    *this << "<title>" << title << "</title>\n";
    *this << "</head>\n\n";
    *this << "<body>\n";
  }
  void outputFooter() {
    *this << "</body>\n";
    *this << "</html>\n";
  }
public:
  HTMLFile( const string& f, const string& t, const string& s = "", const string& charset = "" )
    : ofstream( f.c_str() ), filename( f ), title( t ), stylesheet( s ), charset_string( charset ) {
    if ( *this ) {
      outputHeader();
    }
  }
  virtual ~HTMLFile() {
    if ( *this ) {
      outputFooter();
    }
  }
};

class compare_count {
private:
  FSArray< FSP, int >* count_db;
public:
  compare_count( FSArray< FSP, int >* db ) : count_db( db ) {
  }
  ~compare_count() {}
  bool operator()( FSP f1, FSP f2 ) const {
    int c1 = 0;
    count_db->find( f1, c1 );
    int c2 = 0;
    count_db->find( f2, c2 );
    return c1 > c2;
  }
};

class lexeme_compare_alphabetical {
private:
  machine* mach;
public:
  lexeme_compare_alphabetical( machine* m ) : mach( m ) {
  }
  ~lexeme_compare_alphabetical() {}
  bool operator()( FSP l1, FSP l2 ) const {
    return get_lexeme_name( mach, l1 ) < get_lexeme_name( mach, l2 );
  }
};

class symbol_compare {
public:
  bool operator()( const pair< string, int >& s1, const pair< string, int >& s2 ) const {
    return s1.second > s2.second;
  }
};

//////////////////////////////////////////////////////////////////////

void output_lexeme_data( ostream* ostr, machine* mach, FSP lexeme,
                         FSArray< FSP, int >& lexeme_id_map,
                         FSArray< FSP, int >& lexeme_count_db, FSArray< FSP, pair< int, FSP > >& lexeme_occurrence_db,
                         FSArray< FSP, int >& word_id_map ) {
  int count = 0;
  lexeme_count_db.find( lexeme, count );
  string lexeme_name = get_lexeme_name( mach, lexeme );
  int lexeme_id = 0;
  lexeme_id_map.find( lexeme, lexeme_id );
  *ostr << "No. " << lexeme_id << " <a href=\""
        << get_file_name( "lexeme", lexeme_id / lexeme_file_size ) << "#"
        << lexeme_id << "\">" << " " << lexeme_name << "</a> Freq.=" << count << "\n";
  pair< int, FSP > occurrence;
  if ( lexeme_occurrence_db.find( lexeme, occurrence ) ) {
    Word word( occurrence.second );
    int word_id = get_word_id( mach, word_id_map, word );
    if ( word_id >= 0 ) {
      *ostr << " for <a href=\"" << get_file_name( "lex", word_id / lex_file_size ) << "#"
            << word_id << "\">\"" << get_word_symbol( mach, word ) << "\"</a>\n";
    } else {
      *ostr << " for \"" << get_word_symbol( mach, word ) << "\"";
    }
  }
  return;
}

//////////////////////////////////////////////////////////////////////

void help_message() {
  *error_stream << "Usage: lex2html [options] grammar_module lexicon_name temptable_name lexbank_name output_prefix" << endl;
  *error_stream << "  grammar_module: LiLFeS program to be loaded" << endl;
  *error_stream << "  lexicon_name: lexicon" << endl;
  *error_stream << "  temptable_name: template table" << endl;
  *error_stream << "  lexbank_name: lexbank" << endl;
  *error_stream << "  output_prefix: prefix of output files" << endl;
  *error_stream << "Options:" << endl;
  *error_stream << "  -c file:\tstylesheet" << endl;
  *error_stream << "  -d directory:\toutput directory" << endl;
  *error_stream << "  -lexsize int, -tmpsize int, -lxmsize int, -sntsize int: size of HTML files" << endl;
  *error_stream << "  -charset string: charset specification in HTML header" << endl;
  *error_stream << "  -v:\t\tprint verbose messages" << endl;
  *error_stream << "  -vv:\t\tprint many verbose messages" << endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  char** end = argv + argc;
  ++argv;
  for ( ; argv != end; ++argv ) {
    if ( (*argv)[ 0 ] == '-' ) {
      // option without argument
      if ( std::strcmp( (*argv), "-v" ) == 0 ) {
        verbose_level = 1;
      } else if ( std::strcmp( (*argv), "-vv" ) == 0 ) {
        verbose_level = 2;
      } else {
        // option with argument
        if ( argv + 1 == end ) {
          *error_stream << *argv << " option requires additional argument" << endl;
          return false;
        }
        if ( std::strcmp( (*argv), "-c" ) == 0 ) {
          stylesheet = *(++argv);
        } else if ( std::strcmp( (*argv), "-d" ) == 0 ) {
          output_dir = *(++argv);
        } else if ( std::strcmp( (*argv), "-charset" ) == 0 ) {
          charset_string = *(++argv);

        } else if ( std::strcmp( (*argv), "-lexsize" ) == 0 ) {
          lex_file_size = strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-tmpsize" ) == 0 ) {
          temp_file_size =  strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-lxmsize" ) == 0 ) {
          lexeme_file_size =  strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-sntsize" ) == 0 ) {
          sentence_file_size =  strtol( *(++argv), NULL, 0 );
        } else {
          *error_stream << "Unknown option: " << *argv << endl;
          return false;
        }
      }
    } else {
      // arguments
      if ( end - argv != 5 ) {
        *error_stream << "lex2html requires five arguments" << endl;
        return false;
      }
      grammar_module = argv[ 0 ];
      lexicon_name = argv[ 1 ];
      temptable_name = argv[ 2 ];
      lexbank_name = argv[ 3 ];
      output_prefix = argv[ 4 ];
      return true;
    }
  }
  *error_stream << "lex2html requires five arguments" << endl;
  return false;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

int main( int argc, char** argv ) {

  *error_stream << "lex2html (MAYZ " << MAYZ_VERSION << ")" << endl;
  *error_stream << MAYZ_COPYRIGHT << endl;

  Initializer::PerformAll();

  // show help message
  if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
    help_message();
    return 0;
  }
  // command-line arguments
  if ( ! analyze_arguments( argc, argv ) ) {
    help_message();
    return 1;
  }

  machine mach;

  try {

    // Load grammar module
    *error_stream << "Loading grammar module \"" << grammar_module << "\"... ";
    if ( ! load_module( mach, "grammar module", grammar_module ) ) {
      *error_stream << "Grammar module \"" << grammar_module << "\" not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    // Initialization
    *error_stream << "Initializing... ";
    lexicon_lookup_key = prepare_proc( module::UserModule(), "lexicon_lookup_key", 2 );
    if ( ! lexicon_lookup_key ) {
      *error_stream << "lexicon_lookup_key/2 not found" << endl;
      return 1;
    }
    unknown_word_lookup_key = prepare_proc( module::UserModule(), "unknown_word_lookup_key", 2 );
    if ( ! unknown_word_lookup_key ) {
      *error_stream << "unknown_word_lookup_key/2 not found" << endl;
      return 1;
    }
    lex_template_label = prepare_proc( module::UserModule(), "lex_template_label", 2 );
    if ( ! lex_template_label ) {
      *error_stream << "lex_template_label/2 not found" << endl;
      return 1;
    }
    sign_label = prepare_proc( module::UserModule(), "sign_label", 2 );
    if ( ! sign_label ) {
      *error_stream << "sign_label/2 not found" << endl;
      return 1;
    }
    lexicon_key_label = prepare_proc( module::UserModule(), "lexicon_key_label", 2 );
    if ( ! lexicon_key_label ) {
      *error_stream << "lexicon_key_label/2 not found" << endl;
      return 1;
    }
    word_label = prepare_proc( module::UserModule(), "word_label", 2 );
    if ( ! word_label ) {
      *error_stream << "word_label/2 not found" << endl;
      return 1;
    }

    *error_stream << "done." << endl;

  } catch ( exception& e ) {
    *error_stream << "Initialization failed: " << e.what() << endl;
    return 1;
  }

  //////////////////////////////////////////////////
  //  Load database
  //////////////////////////////////////////////////

  Lexicon lexicon( &mach );
  TemplateTable temptable( &mach );
  Lexbank lexbank( &mach );

  FSArray< int, vector< FSP > > sentence_bank( &mach );    // (sentence id, list of words)
  FSArray< FSP, int > word_count_db( &mach );              // (word, count)
  FSArray< FSP, int > temp_count_db( &mach );              // (template name, count)
  FSArray< FSP, int > lexeme_count_db( &mach );            // (lexeme name, count)
  FSArray< FSP, pair< int, FSP > > temp_occurrence_db( &mach );    // (template name, <sentence id, word>)
  FSArray< FSP, pair< int, FSP > > lexeme_occurrence_db( &mach );  // (lexeme name, <sentence id, word>)
  FSArray< FSP, vector< FSP > > lexeme_expand_db( &mach );         // (lexeme name, list of template names)
  FSArray< FSP, vector< FSP > > lexeme_expand_db2( &mach );        // (lexeme name, list of template names)
  size_t num_sentences = 0;
  int sentence_id = 0;
  int max_sentence_id = 0;

  vector< Word > word_list;
  vector< LexTemplate > temp_list;
  vector< FSP > lexeme_list;
  vector< FSP > lexeme_list_alphabetical;
  FSArray< FSP, int > word_id_map( &mach );      // (word, word id)
  FSArray< FSP, int > temp_id_map( &mach );      // (template name, template id)
  FSArray< FSP, int > lexeme_id_map( &mach );    // (lexeme name, lexeme id)
  FSArray< FSP, string > temp_symbol_db( &mach );  // (template name, NTS)
  map< string, int > symbol_count_db;            // (NTS, count)

  try {
    *error_stream << "Loading lexicon... ";
    if ( ! lexicon.load( lexicon_name ) ) {
      *error_stream << "failed." << endl;
      return 1;
    }
    *error_stream << "done." << endl;
    *error_stream << "Loading template table... ";
    if ( ! temptable.load( temptable_name ) ) {
      *error_stream << "failed." << endl;
      return 1;
    }
    *error_stream << "done." << endl;
    *error_stream << "Loading lexbank... ";
    if ( ! lexbank.load( lexbank_name ) ) {
      *error_stream << "failed." << endl;
      return 1;
    }
    *error_stream << "done." << endl;
  } catch ( exception& e ) {
    *error_stream << "Loading database failed: " << e.what() << endl;
    return 1;
  }

  //////////////////////////////////////////////////
  //  Collect statistics
  //////////////////////////////////////////////////

  *error_stream << "Collecting lexbank statistics..." << endl;
  try {
    for ( Lexbank::iterator sentence_it = lexbank.begin();
          sentence_it != lexbank.end();
          ++sentence_it ) {
      core_p trailp = mach.SetTrailPoint( NULL );
      sentence_id = sentence_it->first;
      ++num_sentences;
      if ( sentence_id > max_sentence_id ) max_sentence_id = sentence_id;
      vector< FSP > sentence;
      for ( vector< FSP >::iterator lex_it = sentence_it->second.begin();
            lex_it != sentence_it->second.end();
            ++lex_it ) {
        Derivation derivterm( *lex_it );
        Word word = derivterm.termWord();
        sentence.push_back( word );
        // increment word count
        FSP lexicon_key = get_lexicon_key( &mach, word );
        inc_count_db( word_count_db, lexicon_key );
        // update template info.
        LexTemplate term_template = derivterm.termTemplate();
        inc_count_db( temp_count_db, term_template );
        update_occurrence_db( temp_occurrence_db, term_template, sentence_id, word );
        // update lexeme info.
        FSP lexeme = term_template.lexemeName();
        inc_count_db( lexeme_count_db, lexeme );
        update_occurrence_db( lexeme_occurrence_db, lexeme, sentence_id, word );
        update_expand_db( lexeme_expand_db2, lexeme, term_template ); 
      }
      if ( ! sentence_bank.insert( sentence_id, sentence ) )
        throw lex2html_exception( "sentence bank update failed" );
      if ( num_sentences % 100 == 0 ) {
        *error_stream << "  " << num_sentences << " sentences processed." << endl;
      }
      mach.TrailBack( trailp );
    }
  } catch ( exception& e ) {
    *error_stream << "Error in sentence " << sentence_id << ": " << e.what() << endl;
    return 1;
  }
  *error_stream << "done." << endl;

  //////////////////////////////////////////////////
  //  Sorting
  //////////////////////////////////////////////////

  try {
    *error_stream << "Sorting lexicon... ";
    for ( Lexicon::iterator word_it = lexicon.begin();
          word_it != lexicon.end();
          ++word_it ) {
      Word word = word_it->first;
      word_list.push_back( word );
    }
    sort( word_list.begin(), word_list.end(), compare_count( &word_count_db ) );
    for ( size_t i = 0; i < word_list.size(); ++i ) {
      word_id_map.insert( word_list[ i ], i );
    }
    *error_stream << "done." << endl;
    *error_stream << "Sorting templates... ";
    for ( TemplateTable::iterator temp_it = temptable.begin();
          temp_it != temptable.end();
          ++temp_it ) {
      LexTemplate temp( temp_it->first );
      temp_list.push_back( temp );
      string symbol = get_sign_symbol( &mach, temp_it->second );
      temp_symbol_db.insert( temp, symbol );
      int temp_count = 0;
      temp_count_db.find( temp, temp_count );
      symbol_count_db[ symbol ] += temp_count;
      update_expand_db( lexeme_expand_db, temp.lexemeName(), temp ); 
    }
    sort( temp_list.begin(), temp_list.end(), compare_count( &temp_count_db ) );
    for ( size_t i = 0; i < temp_list.size(); ++i ) {
      temp_id_map.insert( temp_list[ i ], i );
    }
    *error_stream << "done." << endl;
    *error_stream << "Sorting lexemes... ";
    for ( FSArray< FSP, int >::iterator lexeme_it = lexeme_count_db.begin();
          lexeme_it != lexeme_count_db.end();
          ++lexeme_it ) {
      lexeme_list.push_back( lexeme_it->first );
      lexeme_list_alphabetical.push_back( lexeme_it->first );
    }
    sort( lexeme_list.begin(), lexeme_list.end(), compare_count( &lexeme_count_db ) );
    sort( lexeme_list_alphabetical.begin(), lexeme_list_alphabetical.end(), lexeme_compare_alphabetical( &mach ) );
    for ( size_t i = 0; i < lexeme_list.size(); ++i ) {
      lexeme_id_map.insert( lexeme_list[ i ], i );
    }
    *error_stream << "done." << endl;
  } catch( exception& e ) {
    *error_stream << "Fatal error: " << e.what() << endl;
    return 1;
  }

  //////////////////////////////////////////////////
  //  Output
  //////////////////////////////////////////////////

  *error_stream << "Output HTML files... ";
  try {

    // Output lexicon files
    string output_name = get_output_file_name( "lex", 0 );
    ofstream* output_file = new HTMLFile( output_name.c_str(), "Lexicon", stylesheet, charset_string );
    if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
    int num_lex = 0;
    for ( vector< Word >::iterator word_it = word_list.begin();
          word_it != word_list.end();
          ++word_it ) {
      ++num_lex;
      int count = 0;
      word_count_db.find( *word_it, count );
      int word_id = 0;
      word_id_map.find( *word_it, word_id );
      *output_file << "<div>\n";
      *output_file << "<a name=\"" << word_id << "\">No. " << num_lex << " "
		   << get_lexkey_symbol( &mach, *word_it ) << " Freq.=" << count << "</a>\n";
      vector< FSP > temp_list;
      if ( lexicon.lookup( *word_it, temp_list ) ) {
        *output_file << "<ul>\n";
        for ( vector< FSP >::iterator temp_it = temp_list.begin();
              temp_it != temp_list.end();
              ++temp_it ) {
          LexTemplate temp( *temp_it );
          string template_name = get_template_name( &mach, temp );
          int temp_id = 0;
          temp_id_map.find( temp, temp_id );
          int count = 0;
          temp_count_db.find( temp, count );
          *output_file << "<li><a href=\"" << get_file_name( "temp", temp_id / temp_file_size ) << "#"
                       << temp_id << "\">" << template_name << "</a> Freq.="
                       << count << "\n";
        }
        *output_file << "</ul>\n";
      }
      *output_file << "</div>\n";
      output_index_link( output_file );
      if ( num_lex % lex_file_size == 0 ) {
        delete output_file;
        output_name = get_output_file_name( "lex", num_lex / lex_file_size );
        output_file = new HTMLFile( output_name.c_str(), "Lexicon", stylesheet, charset_string );
        if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
      }
    }
    delete output_file;

    // Output template files
    output_name = get_output_file_name( "temp", 0 );
    output_file = new HTMLFile( output_name.c_str(), "List of templates", stylesheet, charset_string );
    if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
    int num_temp = 0;
    for ( vector< LexTemplate >::iterator temp_it = temp_list.begin();
          temp_it != temp_list.end();
          ++temp_it ) {
      ++num_temp;
      int count = 0;
      temp_count_db.find( *temp_it, count );
      int temp_id = 0;
      temp_id_map.find( *temp_it, temp_id );
      string temp_name = get_template_name( &mach, *temp_it );
      *output_file << "<div><a name=\"" << temp_id << "\">No. " << num_temp << " " << temp_name << " Freq.=" << count << "</a></div>\n";
      pair< int, FSP > occurrence;
      if ( temp_occurrence_db.find( *temp_it, occurrence ) ) {
        Word word( occurrence.second );
        int word_id = get_word_id( &mach, word_id_map, word );
        if ( word_id >= 0 ) {
          *output_file << "<div>Occurred for <a href=\"" << get_file_name( "lex", word_id / lex_file_size )
                       << "#" << word_id << "\">\""
                       << get_word_symbol( &mach, word ) << "\"</a>";
        } else {
          *output_file << "<div>Occurred for \"" << get_word_symbol( &mach, word ) << "\"";
        }
        vector< FSP > dummy_sentence;
        if ( sentence_bank.find( occurrence.first, dummy_sentence ) ) {
          *output_file << " in <a href=\"" << get_file_name( "sentence", occurrence.first / sentence_file_size )
                       << "#" << occurrence.first << "\">sentence " << occurrence.first << "</a></div>\n";
        } else {
          *output_file << " in sentence " << occurrence.first << "</div>\n";
        }
      }
      FSP template_fs( mach );
      if ( temptable.find( *temp_it, template_fs ) ) {
        FSXML fs_xml;
        fs_xml.input( template_fs, "", true );
        *output_file << "<div>" << fs_xml.getString() << "</div>\n";
      }
      output_index_link( output_file );
      if ( num_temp % temp_file_size == 0 ) {
        delete output_file;
        output_name = get_output_file_name( "temp", num_temp / temp_file_size );
        output_file = new HTMLFile( output_name, "List of templates", stylesheet, charset_string );
        if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
      }
    }
    delete output_file;

    // Output lexeme files
    output_name = get_output_file_name( "lexeme", 0 );
    output_file = new HTMLFile( output_name.c_str(), "List of lexemes", stylesheet, charset_string );
    if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
    int num_lexeme = 0;
    for ( vector< FSP >::iterator lexeme_it = lexeme_list.begin();
          lexeme_it != lexeme_list.end();
          ++lexeme_it ) {
      ++num_lexeme;
      int count = 0;
      lexeme_count_db.find( *lexeme_it, count );
      string lexeme_name = get_lexeme_name( &mach, *lexeme_it );
      int lexeme_id = 0;
      lexeme_id_map.find( *lexeme_it, lexeme_id );
      *output_file << "<div><a name=\"" << lexeme_id << "\">No. " << num_lexeme << " " << lexeme_name << " Freq.=" << count << "</a></div>\n";
      pair< int, FSP > occurrence;
      if ( lexeme_occurrence_db.find( *lexeme_it, occurrence ) ) {
        Word word( occurrence.second );
        int word_id = get_word_id( &mach, word_id_map, word );
        if ( word_id >= 0 ) {
          *output_file << "<div>Occurred for <a href=\"" << get_file_name( "lex", word_id / lex_file_size )
                       << "#" << word_id << "\">\""
                       << get_word_symbol( &mach, word ) << "\"</a>";
        } else {
          *output_file << "<div>Occurred for \"" << get_word_symbol( &mach, word ) << "\"";
        }
        vector< FSP > dummy_sentence;
        if ( sentence_bank.find( occurrence.first, dummy_sentence ) ) {
          *output_file << " in <a href=\"" << get_file_name( "sentence", occurrence.first / sentence_file_size )
                       << "#" << occurrence.first << "\">sentence "
                       << occurrence.first << "</a></div>\n";
        } else {
          *output_file << " in sentence " << occurrence.first << "</div>\n";
        }
      }
      vector< FSP > expand_list;
      *output_file << "<ul>\n";
      if ( lexeme_expand_db.find( *lexeme_it, expand_list ) ) {
        stable_sort( expand_list.begin(), expand_list.end(), compare_count( &temp_count_db ) );
        for ( vector< FSP >::iterator expand_it = expand_list.begin();
              expand_it != expand_list.end();
              ++expand_it ) {
          int temp_count = 0;
          temp_count_db.find( *expand_it, temp_count );
          string expand_name = get_template_name( &mach, *expand_it );
          int temp_id = 0;
          temp_id_map.find( *expand_it, temp_id );
          *output_file << "  <li><a href=\"" << get_file_name( "temp", temp_id / temp_file_size )
                       << "#" << temp_id << "\">"
                       << expand_name << "</a> Freq.=" << temp_count << "\n";
        }
      }
      vector< FSP > expand_list2;
      if ( lexeme_expand_db2.find( *lexeme_it, expand_list2 ) ) {
        stable_sort( expand_list2.begin(), expand_list2.end(), compare_count( &temp_count_db ) );
        for ( vector< FSP >::iterator expand_it = expand_list2.begin();
              expand_it != expand_list2.end();
              ++expand_it ) {
          bool flag = false;
          for ( vector< FSP >::iterator it = expand_list.begin();
                it != expand_list.end();
                ++it ) {
            if ( lilfes::builtin::equivalent( mach, *it, *expand_it ) ) {
              flag = true;
              break;
            }
          }
          if ( ! flag ) {
            int temp_count = 0;
            temp_count_db.find( *expand_it, temp_count );
            string expand_name = get_template_name( &mach, *expand_it );
            int temp_id = 0;
            temp_id_map.find( *expand_it, temp_id );
            *output_file << "  <li><a href=\"" << get_file_name( "temp", temp_id / temp_file_size )
                         << "#" << temp_id << "\">"
                         << expand_name << "</a> Freq.=" << temp_count << " (not in the lexicon)\n";
          }
        }
      }
      *output_file << "</ul>\n";
      output_index_link( output_file );
      if ( num_lexeme % lexeme_file_size == 0 ) {
        delete output_file;
        output_name = get_output_file_name( "lexeme", num_lexeme / lexeme_file_size );
        output_file = new HTMLFile( output_name, "List of lexemes", stylesheet, charset_string );
        if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
      }
    }
    delete output_file;

    // Output sentence file
    output_name = get_output_file_name( "sentence", 0 );
    output_file = new HTMLFile( output_name, "List of sentences", stylesheet, charset_string );
    if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
    for ( int i = 1; i < max_sentence_id; ++i ) {
      vector< FSP > sentence;
      if ( sentence_bank.find( i, sentence ) ) {
        *output_file << "<div><a name=\"" << i << "\">No. " << i << ":";
        for ( vector< FSP >::iterator word_it = sentence.begin();
              word_it != sentence.end();
              ++word_it ) {
          Word word( *word_it );
          *output_file << " ";
          int word_id = get_word_id( &mach, word_id_map, word );
          if ( word_id >= 0 ) {
            *output_file << "<a href=\"" << get_file_name( "lex", word_id / lex_file_size )
                         << "#" << word_id << "\">"
                         << get_word_symbol( &mach, word ) << "</a>";
          } else {
            *output_file << get_word_symbol( &mach, word );
          }
        }
        *output_file << "</a></div>\n";
      }
      if ( i % 10 == 0 ) output_index_link( output_file );
      if ( i % sentence_file_size == 0 ) {
        delete output_file;
        output_name = get_output_file_name( "sentence", i / sentence_file_size );
        output_file = new HTMLFile( output_name, "List of sentences", stylesheet, charset_string );
        if ( ! *output_file ) throw lex2html_exception( ( "Cannot open output file: " + output_name ).c_str() );
      }
    }
    delete output_file;

    // Output symbol file
    for ( map< string, int >::const_iterator symbol_it = symbol_count_db.begin();
          symbol_it != symbol_count_db.end();
          ++symbol_it ) {
      string output_symbol_name = get_output_file_name( "symbol-" + symbol_it->first );
      HTMLFile output_symbol_file( output_symbol_name, symbol_it->first, stylesheet, charset_string );
      output_symbol_file << "<h3>" << symbol_it->first << " Freq.=" << symbol_it->second << "</h3>\n";
      // Output in frequency order
      output_symbol_file << "<ol>\n";
      {
        size_t num_lexemes = 0;
        for ( vector< FSP >::iterator lexeme_it = lexeme_list.begin();
              lexeme_it != lexeme_list.end();
              ++lexeme_it ) {
          LexTemplate lex_template( mach );
          lex_template.lexemeName().Unify( *lexeme_it );
          lex_template.lexicalRules().Coerce( nil );
          string symbol;
          if ( temp_symbol_db.find( lex_template, symbol ) &&
               symbol == symbol_it->first ) {
            ++num_lexemes;
            output_symbol_file << "<li>";
            output_lexeme_data( &output_symbol_file, &mach, *lexeme_it,
                                lexeme_id_map, lexeme_count_db, lexeme_occurrence_db,
                                word_id_map );
            if ( num_lexemes % 10 == 0 ) {
              output_symbol_file << "<hr><a href=\"" << output_prefix << "-index.html\">Index</a>\n<hr>\n";
            }
          }
        }
      }
      output_symbol_file << "</ol>\n";
      // Output in alphabetical order
      output_symbol_file << "<h3>Alphabetical order</h3>\n";
      output_symbol_file << "<ol>\n";
      {
        size_t num_lexemes = 0;
        for ( vector< FSP >::iterator lexeme_it = lexeme_list_alphabetical.begin();
              lexeme_it != lexeme_list_alphabetical.end();
              ++lexeme_it ) {
          LexTemplate lex_template( mach );
          lex_template.lexemeName().Unify( *lexeme_it );
          lex_template.lexicalRules().Coerce( nil );
          string symbol;
          if ( temp_symbol_db.find( lex_template, symbol ) &&
               symbol == symbol_it->first ) {
            ++num_lexemes;
            output_symbol_file << "<li>";
            output_lexeme_data( &output_symbol_file, &mach, *lexeme_it,
                                lexeme_id_map, lexeme_count_db, lexeme_occurrence_db,
                                word_id_map );
            if ( num_lexemes % 10 == 0 ) {
              output_symbol_file << "<hr><a href=\"" << output_prefix << "-index.html\">Index</a>\n<hr>\n";
            }
          }
        }
      }
      output_symbol_file << "</ol>\n";
      output_symbol_file << "<hr><a href=\"" << output_prefix << "-index.html\">Index</a>\n<hr>\n";
    }

    // Output index file
    string index_name = get_output_file_name( "index" );
    HTMLFile index_file( index_name, "Index: " + output_prefix, stylesheet, charset_string );
    if ( ! index_file ) throw ( ( "Cannot open output file: " + index_name ).c_str() );
    // sentences
    index_file << "<h3>Index of sentences</h3>\n";
    index_file << "<ul>\n";
    for ( int i = 0; i < max_sentence_id; i += sentence_file_size ) {
      string filename = get_file_name( "sentence", i / sentence_file_size );
      index_file << "  <li><a href=\"" << filename << "\">" << filename << "</a>\n";
    }
    index_file << "</ul>\n";
    // symbols
    index_file << "<h3>Index of symbols</h3>\n";
    index_file << "<ul>\n";
    for ( map< string, int >::iterator it = symbol_count_db.begin();
          it != symbol_count_db.end();
          ++it ) {
      index_file << "  <li><a href=\"" << get_file_name( "symbol-" + it->first ) << "\">" << it->first << "</a> Freq.=" << it->second << "\n";
    }
    index_file << "</ul>\n";
    // lexemes
    index_file << "<h3>Index of lexemes</h3>\n";
    index_file << "<ul>\n";
    for ( size_t i = 0; i < lexeme_list.size(); i += lexeme_file_size ) {
      string filename = get_file_name( "lexeme", i / lexeme_file_size );
      index_file << "  <li><a href=\"" << filename << "\">" << filename << "</a>\n";
    }
    index_file << "</ul>\n";
    // templates
    index_file << "<h3>Index of templates</h3>\n";
    index_file << "<ul>\n";
    for ( size_t i = 0; i < temp_list.size(); i += temp_file_size ) {
      string filename = get_file_name( "temp", i / temp_file_size );
      index_file << "  <li><a href=\"" << filename << "\">" << filename << "</a>\n";
    }
    index_file << "</ul>\n";
    // lexicon
    index_file << "<h3>Index of lexicon</h3>\n";
    index_file << "<ul>\n";
    for ( size_t i = 0; i < word_list.size(); i += lex_file_size ) {
      string filename = get_file_name( "lex", i / lex_file_size );
      index_file << "  <li><a href=\"" << filename << "\">" << filename << "</a>\n";
    }
    index_file << "</ul>\n";
    *error_stream << "done." << endl;

  } catch ( exception& e ) {
    cerr << "Fatal error: " << e.what() << endl;
    return 1;
  }

  Terminator::PerformAll();
}

