/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file lexrefine_main.cc
 *  @version Time-stamp: <2005-06-21 19:24:21 yusuke>
 *  Main program of lexrefine
 *
 **********************************************************************/

#include "Lexicon.h"
#include "TemplateTable.h"
#include "Initializer.h"
#include <string>
#include <iostream>
#include <set>

using namespace std;
using namespace mayz;
using namespace lilfes;

static int word_frequency_threshold = 1;
static int template_frequency_threshold = 0;
static int unknown_word_frequency_threshold = 1;
static int unknown_word_template_frequency_threshold = 0;
//static int expand_template_threshold = 0;
static int verbose_level = 0;

static string rule_file_name;
static string orig_lexicon_file;
static string orig_template_file;
static string new_lexicon_file;
static string new_template_file;

procedure* unknown_word_key = NULL;
procedure* expand_lexical_template = NULL;
procedure* expand_lexicon = NULL;

//////////////////////////////////////////////////////////////////////

void help_message() {
  *error_stream << "Usage: lexrefine [options] lexrule_module orig_lexicon_file orig_template_file new_lexicon_file new_template_file" << endl;
  *error_stream << "  lexrule_module: LiLFeS module in which lexical rules are defined" << endl;
  *error_stream << "  orig_lexicon_file: input lexicon database" << endl;
  *error_stream << "  orig_template_file: input template database" << endl;
  *error_stream << "  new_lexicon_file: output lexicon database" << endl;
  *error_stream << "  new_template_file: output template database" << endl;
  *error_stream << "Options:" << endl;
  *error_stream << "  -wf threshold:\tword frequency threshold (default: " << word_frequency_threshold << ")" << endl;
  *error_stream << "  -tf threshold:\ttemplate frequency threshold (default: " << template_frequency_threshold << ")" << endl;
  *error_stream << "  -uwf threshold:\tfrequency threshold of unknown words (default: " << unknown_word_frequency_threshold << ")" << endl;
  *error_stream << "  -utf threshold:\tfrequency threshold of unknown word templates (default: " << unknown_word_template_frequency_threshold << ")" << endl;
  //  *error_stream << "  -lf\tfrequency threshold of lexemes to be expanded (default: " << expand_template_threshold << ")" << endl;
  *error_stream << "  -v:\tprint verbose messages" << endl;
  *error_stream << "  -vv:\tprint many verbose messages" << endl;
  *error_stream << "  -vvv:\tprint many many verbose messages" << endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  char** end = argv + argc;
  ++argv;
  for ( ; argv != end; ++argv ) {
    if ( (*argv)[ 0 ] == '-' ) {
      // option without argument
      if ( std::strcmp( (*argv), "-v" ) == 0 ) {
        verbose_level = 1;
      } else if ( std::strcmp( (*argv), "-vv" ) == 0 ) {
        verbose_level = 2;
      } else if ( std::strcmp( (*argv), "-vvv" ) == 0 ) {
        verbose_level = 3;
      } else {
        // option with argument
        if ( argv + 1 == end ) {
          *error_stream << *argv << " option requires additional argument" << endl;
          return false;
        }
        if ( std::strcmp( (*argv), "-wf" ) == 0 ) {
          word_frequency_threshold = strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-tf" ) == 0 ) {
          template_frequency_threshold = strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-uwf" ) == 0 ) {
          unknown_word_frequency_threshold = strtol( *(++argv), NULL, 0 );
        } else if ( std::strcmp( (*argv), "-utf" ) == 0 ) {
          unknown_word_template_frequency_threshold = strtol( *(++argv), NULL, 0 );
//          } else if ( std::strcmp( (*argv), "-lf" ) == 0 ) {
//            expand_template_threshold = strtol( *(++argv), NULL, 0 );
        } else {
          *error_stream << "Unknown option: " << *argv << endl;
          return false;
        }
      }
    } else {
      // arguments
      if ( end - argv != 5 ) {
        *error_stream << "lexrefine requires five arguments" << endl;
        return false;
      }
      rule_file_name = argv[ 0 ];
      orig_lexicon_file = argv[ 1 ];
      orig_template_file = argv[ 2 ];
      new_lexicon_file = argv[ 3 ];
      new_template_file = argv[ 4 ];
      return true;
    }
  }
  *error_stream << "lexrefine requires five arguments" << endl;
  return false;
}

//////////////////////////////////////////////////////////////////////

bool apply_unknown_word_key( machine* mach, FSP word, FSP unknown_word ) {
  std::vector< FSP > args( 2 );
  args[ 0 ] = word;
  args[ 1 ] = unknown_word;  // lexicon key of unknown word
  core_p cutp = mach->GetCutPoint();
  if ( ! call_proc( *mach, unknown_word_key, args ) ) return false;
  mach->DoCut( cutp );
  return true;
}

bool apply_expand_lexicon( machine* mach, FSP orig_key, int count, vector< FSP >& new_temp_list, Lexicon* expanded_lexicon, FSArray< FSP, int >* lex_count_db ) {
  if ( expand_lexicon ) {
    vector< cell > tmp_fs;
    orig_key.Serialize( tmp_fs );
    set< vector< cell > > new_key_list;
    new_key_list.insert( tmp_fs );
    for ( vector< FSP >::iterator new_temp_it = new_temp_list.begin();
          new_temp_it != new_temp_list.end();
          ++new_temp_it ) {
      if ( verbose_level > 1 ) {
        *error_stream << "expanding for:" << endl;
        *error_stream << new_temp_it->DisplayAVM() << flush;
      }
      vector< FSP > args( 3 );
      args[ 0 ] = orig_key;
      args[ 1 ] = *new_temp_it;
      args[ 2 ] = FSP( mach );    // new key
      vector< bool > flags( 3, false );
      flags[ 2 ] = true;
      list< vector< FSP > > results;
      findall_proc( *mach, expand_lexicon, args, flags, results );
      for ( list< vector< FSP > >::iterator result_it = results.begin();
            result_it != results.end();
            ++result_it ) {
        // add new template to new key
        vector< FSP > temp_list;
        expanded_lexicon->lookup( (*result_it)[ 0 ], temp_list );
        bool already_exist = false;
        for ( vector< FSP >::iterator temp_it = temp_list.begin();
              temp_it != temp_list.end();
              ++temp_it ) {
          if ( builtin::equivalent( *mach, *new_temp_it, *temp_it ) ) {
            already_exist = true;
            break;
          }
        }
        if ( ! already_exist ) {
          temp_list.push_back( *new_temp_it );
          if ( ! expanded_lexicon->insert( (*result_it)[ 0 ], temp_list ) ) {
            *error_stream << "database access failed: expand lexicon" << endl;
            return false;
          }
          if ( verbose_level > 1 ) {
            *error_stream << "expanded word:" << endl;
            *error_stream << (*result_it)[ 0 ].DisplayAVM() << flush;
          }
        }
        tmp_fs.resize( 0 );
        (*result_it)[ 0 ].Serialize( tmp_fs );
        if ( new_key_list.insert( tmp_fs ).second ) {
          int new_key_count = 0;
          lex_count_db->find( (*result_it)[ 0 ], new_key_count );
          new_key_count += count;
          if ( ! lex_count_db->insert( (*result_it)[ 0 ], new_key_count ) ) {
            *error_stream << "database access failed: new key count" << endl;
            return false;
          }
          if ( verbose_level > 1 ) {
            *error_stream << "update count: " << new_key_count << endl;
          }
        }
      }
    }
  }
  return true;
}

//////////////////////////////////////////////////////////////////////

int main( int argc, char** argv ) {

  *error_stream << "lexrefine (MAYZ " << MAYZ_VERSION << ")" << endl;
  *error_stream << MAYZ_COPYRIGHT << endl;

  lilfes::Initializer::PerformAll();

  // show help message
  if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
    help_message();
    return 0;
  }
  // command-line arguments
  if ( ! analyze_arguments( argc, argv ) ) {
    help_message();
    return 1;
  }

  machine mach;

  try {
    // Load rule file
    *error_stream << "Loading rule file \"" << rule_file_name << "\"... ";
    if ( ! load_module( mach, "lexicon refine module", rule_file_name ) ) {
      *error_stream << "\nRule file \"" << rule_file_name << "\" not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    // Initialization
    *error_stream << "Initializing... ";
    if ( ! mayz::Initializer::initialize( &mach ) ) {
      *error_stream << "initialization failed" << endl;
      return 1;
    }
    unknown_word_key = prepare_proc( module::UserModule(), "unknown_word_key", 2 );
    if ( ! unknown_word_key ) {
      *error_stream << "'unknown_word_key/2' not found" << endl;
      return 1;
    }
    expand_lexical_template = prepare_proc( module::UserModule(), "expand_lexical_template", 5 );
    expand_lexicon = prepare_proc( module::UserModule(), "expand_lexicon", 3 );
    *error_stream << "done." << endl;
  } catch ( exception& e ) {
    *error_stream << "Initialization failed: " << e.what() << endl;
    return 1;
  }

  try {

    // Load original lexicon
    *error_stream << "Loading original lexicon and templates... ";
    Lexicon orig_lexicon( &mach );
    TemplateTable orig_temptable( &mach );
    Lexicon new_lexicon( &mach );
    TemplateTable new_temptable( &mach );
    FSArray< FSP, int > lex_count_db( &mach );
    if ( ! orig_lexicon.load( orig_lexicon_file ) ) {
      *error_stream << "Loading lexicon failed" << endl;
      return 1;
    }
    if ( ! orig_temptable.load( orig_template_file, orig_template_file + ".count" ) ) {
      *error_stream << "Loading template database failed" << endl;
      return 1;
    }
    if ( ! lex_count_db.load( orig_lexicon_file + ".count" ) ) {
      *error_stream << "Loading lexicon count database failed" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    //////////////////////////////////////////////////////////////////////
    ////  Stage 1: Refine template table
    //////////////////////////////////////////////////////////////////////

    *error_stream << "Refine template table..." << endl;
    int num_processed_templates = 0;
    int num_original_templates = 0;
    int num_expanded_templates = 0;
    FSArray< FSP, vector< FSP > > template_mapping( &mach );
    for ( TemplateTable::iterator temp_it = orig_temptable.begin();
          temp_it != orig_temptable.end();
          ++temp_it ) {
      core_p trailp = mach.SetTrailPoint( NULL );
      ++num_processed_templates;
      int count = 0;
      if ( ! orig_temptable.getCount( temp_it->first, count ) ) {
        *error_stream << "Frequency of a template not found" << endl;
        return 1;
      }
      if ( verbose_level > 1 ) {
        *error_stream << "processing template: count=" << count << endl;
        *error_stream << temp_it->first.DisplayAVM() << flush;
      }
      if ( count > template_frequency_threshold ) {
        // more than the threshold
        //   -> register to the new table
        ++num_original_templates;
        if ( ! new_temptable.insert( temp_it->first, temp_it->second ) ) {
          *error_stream << "database access error: insert in refining template table" << endl;
          return 1;
        }
        if ( ! new_temptable.incCount( temp_it->first, count ) ) {
          *error_stream << "database access error: inc count in refining template table" << endl;
          return 1;
        }
        if ( verbose_level > 1 ) *error_stream << "registered to the template table" << endl;
        // expansion by lexical rules
        if ( expand_lexical_template ) {
          LexTemplate orig_temp_name( temp_it->first );
          vector< FSP > args( 5 );
          args[ 0 ] = temp_it->first;
          args[ 1 ] = temp_it->second;
          args[ 2 ] = FSP( mach, count );
          args[ 3 ] = FSP( mach );    // list of lexical rules
          args[ 4 ] = FSP( mach );    // new template
          vector< bool > flags( 5, false );
          flags[ 3 ] = true;
          flags[ 4 ] = true;
          list< vector< FSP > > results;
          vector< FSP > new_templates;
          findall_proc( mach, expand_lexical_template, args, flags, results );
          for ( list< vector< FSP > >::iterator result_it = results.begin();
                result_it != results.end();
                ++result_it ) {
            ++num_expanded_templates;
            LexTemplate new_temp_name( mach );
            new_temp_name.lexemeName().Unify( orig_temp_name.lexemeName() );
            new_temp_name.lexicalRules().Unify( (*result_it)[ 0 ] );
            if ( ! new_temptable.insert( new_temp_name, (*result_it)[ 1 ] ) ) {
              *error_stream << "database access error: insert in template expansion" << endl;
              return 1;
            }
            if ( ! new_temptable.incCount( new_temp_name, count ) ) {
              *error_stream << "database access error: inc count in template expansion" << endl;
              return 1;
            }
            if ( verbose_level > 1 ) {
              *error_stream << "expanded to:" << endl;
              *error_stream << new_temp_name.DisplayAVM() << flush;
            }
            if ( verbose_level > 2 ) {
              *error_stream << (*result_it)[ 1 ].DisplayAVM() << flush;
            }
            new_templates.push_back( new_temp_name );
          }
          if ( ! template_mapping.insert( orig_temp_name, new_templates ) ) {
            *error_stream << "database access error: insert in updating temlate mapping" << endl;
            return 1;
          }
        } // if ( expand_lexical_template )
      } else {
        // less than the threshold
        //   -> removed from the table
        if ( verbose_level > 1 ) *error_stream << "removed from the template table" << endl;
      }
      mach.TrailBack( trailp );
      // profiling message
      if ( verbose_level > 0 || num_processed_templates % 100 == 0 ) {
        *error_stream << num_processed_templates << " templates processed" << endl;
        *error_stream << "  Number of original templates = " << num_original_templates << endl;
        *error_stream << "  Number of expanded templates = " << num_expanded_templates << endl;
      }
    }
    *error_stream << "done." << endl;
    *error_stream << num_processed_templates << " templates processed" << endl;
    *error_stream << "  Number of original templates = " << num_original_templates << endl;
    *error_stream << "  Number of expanded templates = " << num_expanded_templates << endl;

    //////////////////////////////////////////////////////////////////////
    //// Stage 2: Expand lexicon
    //////////////////////////////////////////////////////////////////////

    *error_stream << "Expand lexicon..." << endl;
    int num_processed_words = 0;
    Lexicon expanded_lexicon( &mach );
    for ( Lexicon::iterator word_it = orig_lexicon.begin();
          word_it != orig_lexicon.end();
          ++word_it ) {
      core_p trailp = mach.SetTrailPoint( NULL );
      ++num_processed_words;
      // get word count
      int count = 0;
      //*error_stream << "a2" << endl;
      if ( ! lex_count_db.find( word_it->first, count ) ) {
        *error_stream << "Frequency of words not found" << endl;
        return 1;
      }
      if ( verbose_level > 1 ) {
        *error_stream << "processing word: count=" << count << endl;
        *error_stream << word_it->first.DisplayAVM() << flush;
      }
      for ( vector< FSP >::iterator temp_it = word_it->second.begin();
            temp_it != word_it->second.end();
            ++temp_it ) {
        FSP fs( mach );
        if ( new_temptable.find( *temp_it, fs ) ) {
          // valid template
          if ( verbose_level > 1 ) {
            *error_stream << "expanding template:" << endl;
            *error_stream << temp_it->DisplayAVM() << flush;
          }
          if ( ! expanded_lexicon.add( word_it->first, *temp_it ) ) {
            *error_stream << "add to expanded lexicon failed" << endl;
            return 1;
          }
          vector< FSP > new_temp_list;
          if ( template_mapping.find( *temp_it, new_temp_list ) ) {
            // expansion
            if ( ! apply_expand_lexicon( &mach, word_it->first, count, new_temp_list, &expanded_lexicon, &lex_count_db ) ) {
              *error_stream << "expand_lexicon failed" << endl;
              return 1;
            }
          }
        } else {
          if ( verbose_level > 1 ) {
            *error_stream << "removed template:" << endl;
            *error_stream << temp_it->DisplayAVM() << flush;
          }
        }
      }
      mach.TrailBack( trailp );
      if ( verbose_level > 0 || num_processed_words % 1000 == 0 ) {
        *error_stream << num_processed_words << " words processed" << endl;
      }
    }
    *error_stream << "done." << endl;
    *error_stream << num_processed_words << " words processed" << endl;

    //////////////////////////////////////////////////////////////////////
    //// Stage 3: Refine lexicon
    //////////////////////////////////////////////////////////////////////

    *error_stream << "Refine lexicon..." << endl;
    num_processed_words = 0;
    int num_original_words = 0;
    for ( Lexicon::iterator word_it = expanded_lexicon.begin();
          word_it != expanded_lexicon.end();
          ++word_it ) {
      core_p trailp = mach.SetTrailPoint( NULL );
      ++num_processed_words;
      int count = 0;
      //*error_stream << "a2" << endl;
      if ( ! lex_count_db.find( word_it->first, count ) ) {
        //*error_stream << "Frequency of words not found" << endl;
        //return 1;
        // maybe unknown word entry
        continue;
      }
      if ( verbose_level > 1 ) {
        *error_stream << "processing word: count=" << count << endl;
        *error_stream << word_it->first.DisplayAVM() << flush;
      }
      if ( count <= unknown_word_frequency_threshold ) {
        // register as an unknown word
        FSP unknown_word( mach );
        if ( ! apply_unknown_word_key( &mach, word_it->first, unknown_word ) ) {
          *error_stream << "cannot call 'unknown_word_key/2'" << endl;
          return 1;
        }
        vector< FSP > temp_id_list;
        expanded_lexicon.lookup( unknown_word, temp_id_list );
        if ( verbose_level > 1 ) {
          *error_stream << "registered as an unknown word:" << endl;
          *error_stream << unknown_word.DisplayAVM();
        }
        //*error_stream << "a9" << endl;
        bool dirty = false;
        for ( vector< FSP >::iterator temp_it = word_it->second.begin();
              temp_it != word_it->second.end();
              ++temp_it ) {
          int count = 0;
          if ( ! new_temptable.getCount( *temp_it, count ) ) {
            *error_stream << "Frequency of a template not found" << endl;
            return 1;
          }
          if ( count > unknown_word_template_frequency_threshold ) {
            bool already_exist = false;
            for ( vector< FSP >::iterator temp_it2 = temp_id_list.begin();
                  temp_it2 != temp_id_list.end();
                  ++temp_it2 ) {
              if ( builtin::equivalent( mach, *temp_it, *temp_it2 ) ) {
                already_exist = true;
                break;
              }
            }
            if ( ! already_exist ) {
              // new template for unknown word
              if ( verbose_level > 1 ) {
                *error_stream << "template added:" << endl;
                *error_stream << temp_it->DisplayAVM() << flush;
              }
              temp_id_list.push_back( *temp_it );
              dirty = true;
            } else {
              if ( verbose_level > 1 ) {
                *error_stream << "template already exists:" << endl;
                *error_stream << temp_it->DisplayAVM() << flush;
              }
            }
          } else {
            if ( verbose_level > 1 ) {
              *error_stream << "template ignored:" << endl;
              *error_stream << temp_it->DisplayAVM() << flush;
            }
          }
        }
        if ( dirty ) {
          if ( ! expanded_lexicon.insert( unknown_word, temp_id_list ) ) {
            *error_stream << "database access error: database update for unknown word" << endl;
            return 1;
          }
        }
        //*error_stream << "a11" << endl;
        //*error_stream << unknown_word.DisplayAVM();
        //*error_stream << "a12" << endl;
      }
      if ( count > word_frequency_threshold ) {
        // frequent word
        ++num_original_words;
        vector< FSP > temp_id_list;
        //*error_stream << "a4" << endl;
        if ( verbose_level > 1 ) *error_stream << "registered to the lexicon" << endl;
        for ( vector< FSP >::iterator temp_it = word_it->second.begin();
              temp_it != word_it->second.end();
              ++temp_it ) {
          FSP fs( mach );
          if ( new_temptable.find( *temp_it, fs ) ) {
            // valid template
            if ( verbose_level > 1 ) {
              *error_stream << "template added:" << endl;
              *error_stream << temp_it->DisplayAVM() << flush;
            }
            temp_id_list.push_back( *temp_it );
          }
        }
        //*error_stream << "a5" << endl;
        if ( ! expanded_lexicon.insert( word_it->first, temp_id_list ) ) {
          *error_stream << "database access error: update original lexicon" << endl;
          return 1;
        }
        //*error_stream << "a6" << endl;
      } else {
        if ( verbose_level > 1 ) *error_stream << "removed from the lexicon" << endl;
        vector< FSP > dummy_temp_list;
        if ( ! expanded_lexicon.insert( word_it->first, dummy_temp_list ) ) {
          *error_stream << "database access error: deletion from the lexicon" << endl;
          return 1;
        }
      }
      mach.TrailBack( trailp );
      //*error_stream << "a13" << endl;
      if ( verbose_level > 0 || num_processed_words % 1000 == 0 ) {
        *error_stream << num_processed_words << " words processed" << endl;
        *error_stream << "  Number of original words = " << num_original_words << endl;
      }
    }
    *error_stream << "done." << endl;
    *error_stream << num_processed_words << " words processed" << endl;
    *error_stream << "  Number of original words = " << num_original_words << endl;

    //////////////////////////////////////////////////////////////////////
    ////  Stage 4: Make new lexicon
    //////////////////////////////////////////////////////////////////////

    *error_stream << "Making new lexicon..." << endl;
    for ( Lexicon::iterator word_it = expanded_lexicon.begin();
          word_it != expanded_lexicon.end();
          ++word_it ) {
      vector< FSP >& temp_list = word_it->second;
      if ( temp_list.empty() ) continue;  // deleted entry
      if ( ! new_lexicon.insert( word_it->first, word_it->second ) ) {
        *error_stream << "database access error: insertion in new lexicon" << endl;
        return 1;
      }
    }
    *error_stream << "done." << endl;

    // Save new lexicon
    *error_stream << "Output refined lexicon... ";
    new_lexicon.save( new_lexicon_file );
    new_temptable.save( new_template_file );
    *error_stream << "done." << endl;

  } catch ( exception& e ) {
    cerr << "Fatal error: " << e.what() << endl;
    return 1;
  }

  Terminator::PerformAll();

  return 0;
}

