/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file mayzup_main.cc
 *  @version Time-stamp: <2005-06-21 19:31:07 yusuke>
 *  Main program of mayzup
 *
 **********************************************************************/

#include <liblilfes/lilfes.h>
#include <liblilfes/builtin.h>
#include <liblilfes/utility.h>
#include <liblilfes/option.h>
#include <liblilfes/rl_stream.h>
#include "ParserCkyIter.h"
#include "AmisModel.h"

using namespace lilfes;

static const std::string LILFES_PATH_ENV( "LILFES_PATH" );
static const std::string PARSER_NAME( "mayz/up" );

static const int BEAM_NUM_START = 12;
static const double BEAM_WIDTH_START = 6.0;
static const int BEAM_NUM_END = 30;
static const double BEAM_WIDTH_END = 15.0;
static const int BEAM_NUM_STEP = 6;
static const double BEAM_WIDTH_STEP = 3.0;

static std::vector< std::string > module_list;
static std::vector< std::string > command_list;
static bool is_interactive = false;
static size_t limit_sentence_length = 0;
static size_t limit_edge_number = 0;

enum ParserType {
  ITER, FOM, NOFOM
};
static ParserType parser_type = ITER;

void help_message() {
  *error_stream << "Usage: mayzup [options] [-a arguments]" << std::endl;
  *error_stream << "Arguments following \"-a\" are passed to LiLFeS programs as command-line arguments" << std::endl;
  *error_stream << "Options:" << std::endl;
  *error_stream << "  -L dir\tDirectory of LiLFeS modules" << std::endl;
  *error_stream << "  -W number\tLimit number of words" << std::endl;
  *error_stream << "  -E number\tLimit number of edges" << std::endl;
  *error_stream << "  -l module\tLoad LiLFeS program" << std::endl;
  *error_stream << "  -e command\tExecute LiLFeS command" << std::endl;
  *error_stream << "  -i\t\tInteractive mode (show LiLFeS prompt)" << std::endl;
  *error_stream << "  -n\t\tNon-interactive mode (default)" << std::endl;
  *error_stream << "  -iter\t\tWith iterative beam thresholding (default)" << std::endl;
  *error_stream << "  -fom\t\tWith disambiguation" << std::endl;
  *error_stream << "  -nofom\tWithout disambiguation" << std::endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  for ( int i = 1; i < argc; ++i ) {
    if ( argv[ i ][ 0 ] == '-' ) {
      // "-a" option
      if ( std::strcmp( argv[ i ], "-a" ) == 0 ) {
        lilfes_arg_flag::SetFlagFromStringArray( argc - i - 1, argv + i + 1 );
        break;
      }

      // option without argument
      if ( std::strcmp( argv[ i ], "-i" ) == 0 ) {
        // interactive mode
        is_interactive = true;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-n" ) == 0 ) {
        // interactive mode
        is_interactive = false;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-iter" ) == 0 ) {
        // iterative beam thresholding
        parser_type = ITER;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-fom" ) == 0 ) {
        // beam thresholding
        parser_type = FOM;
        continue;
      }
      if ( std::strcmp( argv[ i ], "-nofom" ) == 0 ) {
        // without disambiguation
        parser_type = NOFOM;
        continue;
      }

      // option with argument
      if ( i + 1 < argc ) {
        if ( std::strcmp( argv[ i ], "-L" ) == 0 ) {
          // data path
          const char* lilfes_path = getenv( LILFES_PATH_ENV.c_str() );
          if ( lilfes_path == NULL || *lilfes_path == '\0' ) {
            setenv( LILFES_PATH_ENV.c_str(), argv[ ++i ], 1 );
          } else {
            std::string lilfes_path_str( argv[ ++i ] );
            lilfes_path_str += ':';
            lilfes_path_str += lilfes_path;
            setenv( LILFES_PATH_ENV.c_str(), lilfes_path_str.c_str(), 1 );
          }
          continue;
        }
        if ( std::strcmp( argv[ i ], "-W" ) == 0 ) {
          // limit sentence length
          char* endptr;
          int len = strtol( argv[ ++i ], &endptr, 0 );
          if ( *endptr != '\0' || len < 0 ) {
            *error_stream << "argument of \"-W\" must be a positive integer" << std::endl;
            return false;
          }
          limit_sentence_length = len;
          continue;
        }
        if ( std::strcmp( argv[ i ], "-E" ) == 0 ) {
          // limit edge number
          char* endptr;
          int num = strtol( argv[ ++i ], &endptr, 0 );
          if ( *endptr != '\0' || num < 0 ) {
            *error_stream << "argument of \"-E\" must be a positive integer" << std::endl;
            return false;
          }
          limit_edge_number = num;
          continue;
        }
        if ( std::strcmp( argv[ i ], "-l" ) == 0 ) {
          // lilfes module
          module_list.push_back( argv[ ++i ] );
          continue;
        }
        if ( std::strcmp( argv[ i ], "-e" ) == 0 ) {
          // lilfes command
          command_list.push_back( argv[ ++i ] );
          continue;
        }
      }
      *error_stream << "Unknown option: " << argv[ i ] << std::endl;
      return false;
    } else {
      *error_stream << "Unknown argument: " << argv[ i ] << std::endl;
      return false;
    }
  }
  return true;
}

int main( int argc, char** argv ) {

  *error_stream << "mayzup (MAYZ " << MAYZ_VERSION << ")" << std::endl;
  *error_stream << MAYZ_COPYRIGHT << std::endl;

  try {
    // Initialize everything
    Initializer::PerformAll();

    // Initialize a machine object
    machine mach;

    // show help message
    if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
      help_message();
      return 0;
    }
    // command-line arguments
    if ( ! analyze_arguments( argc, argv ) ) {
      help_message();
      return 1;
    }

    // Load modules
    for ( std::vector< std::string >::const_iterator it = module_list.begin();
          it != module_list.end();
          ++it ) {
      *error_stream << "Loading module \"" << *it << "\"... ";
      if ( ! load_module( mach, "lilfes module", *it ) ) {
        *error_stream << "not found" << std::endl;
        return 1;
      }
      *error_stream << "done." << std::endl;
    }

    // Make a parser
    *error_stream << "Initializing parser..." << std::endl;
    if ( ! load_module( mach, "parser module", PARSER_NAME ) ) {
      *error_stream << "Parser module \"" << PARSER_NAME << "\" not found" << std::endl;
      return 1;
    }
    up::Grammar grammar;
    if(! grammar.init( &mach ) ) {
      *error_stream << "initialization of a grammar failed." << std::endl;
      return 1;
    }
    up::ParserCky* parser = NULL;
    switch ( parser_type ) {
    case ITER: {
      up::ParserCkyIter* p = new up::ParserCkyIter;
      p->enableIterativeParse( BEAM_NUM_START, BEAM_WIDTH_START,
                               BEAM_NUM_END, BEAM_WIDTH_END,
                               BEAM_NUM_STEP, BEAM_WIDTH_STEP );
      parser = p;
      break;
    }
    case FOM: {
      up::ParserCkyFom* p = new up::ParserCkyFom;
      p->setBeamThresholdNum(BEAM_NUM_START);
      p->setBeamThresholdWidth(BEAM_WIDTH_START);
      parser = p;
      break;
    }
    case NOFOM:
      parser = new up::ParserCky;
      break;
    }
    if(! parser->init( &mach, &grammar ) ) {
      *error_stream << "failed." << std::endl;
      return 1;
    }
    if ( limit_sentence_length > 0 ) {
      parser->setLimitWords( limit_sentence_length );
    }
    if ( limit_edge_number > 0 ) {
      parser->setLimitEdges( limit_edge_number );
    }
    parser->enableQuickCheck();
    *error_stream << "done." << std::endl;

    // Start lilfes command
    module::SetCurrentModule( module::UserModule() );
    for ( std::vector< std::string >::iterator it = command_list.begin();
          it != command_list.end();
          ++it ) {
      if ( (*it)[ it->find_last_not_of( " \t" ) ] != '.' ) {
        it->append( "." );
      }
      FSP command = mach.eval( it->c_str() );
      if ( ! command.IsValid() ) {
        *error_stream << "Command not found: " << *it << std::endl;
        return 1;
      }
      command = command.Follow( hd );  // predicate is in hd:
      if ( ! mach.call( command ) ) {
        *error_stream << "Execution of command failed: " << *it << std::endl;
      }
    }

    // Start interactive mode
    if ( is_interactive ) {
      lilfes::lexer_t lexer(&lilfes::rl_istream::rl_stdin, "stdin");
      mach.parse(lexer);
    }

    // Termination process
    if(! parser->term()) {
        *error_stream << "parser failed to terminate" << std::endl;
        return 1;
    }
    if(! grammar.term()) {
      *error_stream << "termination of the grammar failed" << std::endl;
      return 1;
    }
    Terminator::PerformAll();

    return 0;
  }
  catch (std::bad_alloc) {
    std::cerr << "Memory allocation failed, LiLFeS terminated." << std::endl;
    throw;
  }
  catch (std::exception &e) {
    std::cerr << "Fatal error: " << e.what() << std::endl;
    throw;
  }
}

