/**********************************************************************
 *
 *  Copyright (c) 2005, Tsujii Laboratory, The University of Tokyo.
 *  All rights reserved.
 *
 *  @file unimaker_main.cc
 *  @version Time-stamp: <2005-06-21 19:25:10 yusuke>
 *  Main program of unimaker
 *
 **********************************************************************/

#include <liblilfes/gfstream.h>
#include "AmisModel.h"
#include "Derivation.h"
#include "Lexicon.h"
#include "Lexbank.h"
#include "Initializer.h"

using namespace std;
using namespace lilfes;
using namespace mayz;

static string model_name;
static string feature_file_name;
static string lexicon_file_name;
static string template_file_name;
static string lexbank_file_name;
static string uevent_file_name;

static procedure* extract_lexical_event = NULL;
static procedure* extract_lexical_event_feature_value = NULL;
static procedure* import_lexicon = NULL;
static procedure* lookup_lexicon = NULL;

static bool feature_forest_mode = false;
static int limit_sentences = 0;
static int verbose_level = 0;

//////////////////////////////////////////////////////////////////////

class UniMakerException : public std::exception {
private:
  std::string message;
public:
  UniMakerException( const char* m ) : message( m ) {}
  ~UniMakerException() throw () {}
  const char* what() const throw () { return message.c_str(); }
};

//////////////////////////////////////////////////////////////////////

void help_message() {
  *error_stream << "Usage: unimaker model_name lilfes_module lexicon_file template_file lexbank_file uevent_file" << endl;
  *error_stream << "  model_name: name of the model to be made" << endl;
  *error_stream << "  lilfes_module: LiLFeS module in which event extraction predicates are defined" << endl;
  *error_stream << "  lexicon_file: lexicon database file" << endl;
  *error_stream << "  template_file: template database file" << endl;
  *error_stream << "  lexbank_file: lexbank file" << endl;
  *error_stream << "  uevent_file: output file " << endl;
  *error_stream << "Options:" << endl;
  *error_stream << "  -ff:\t\toutput in feature forest format" << endl;
  *error_stream << "  -n num:\tlimit number of sentences (default: " << limit_sentences << ")" << endl;
  *error_stream << "  -v:\t\tprint verbose messages" << endl;
  *error_stream << "  -vv:\t\tprint many verbose messages" << endl;
  return;
}

bool analyze_arguments( int argc, char** argv ) {
  char** end = argv + argc;
  ++argv;
  for ( ; argv != end; ++argv ) {
    if ( (*argv)[ 0 ] == '-' ) {
      // option without argument
      if ( std::strcmp( (*argv), "-v" ) == 0 ) {
        verbose_level = 1;
      } else if ( std::strcmp( (*argv), "-vv" ) == 0 ) {
        verbose_level = 2;
      } else if ( std::strcmp( (*argv), "-ff" ) == 0 ) {
        feature_forest_mode = true;
      } else {
        // option with argument
        if ( argv + 1 == end ) {
          *error_stream << *argv << " option requires additional argument" << endl;
          return false;
        }
        if ( std::strcmp( (*argv), "-n" ) == 0 ) {
          limit_sentences = strtol( *(++argv), NULL, 0 );
        } else {
          *error_stream << "Unknown option: " << *argv << endl;
          return false;
        }
      }
    } else {
      // arguments
      if ( end - argv != 6 ) {
        *error_stream << "unimaker requires six arguments" << endl;
        return false;
      }
      model_name = argv[ 0 ];
      feature_file_name = argv[ 1 ];
      lexicon_file_name = argv[ 2 ];
      template_file_name = argv[ 3 ];
      lexbank_file_name = argv[ 4 ];
      uevent_file_name = argv[ 5 ];
      return true;
    }
  }
  *error_stream << "unimaker requires six arguments" << endl;
  return false;
}

//////////////////////////////////////////////////////////////////////

// extract events
void extract_events( machine& mach, LexEntry lexentry, vector< string >& events ) {
  if ( extract_lexical_event ) {
    vector< FSP > args( 4 );
    args[ 0 ] = FSP( mach, model_name.c_str() );  // model name
    args[ 1 ] = FSP( mach );  // category
    args[ 2 ] = lexentry;     // lexical entry
    args[ 3 ] = FSP( mach );  // event (list of strings)
    vector< bool > flags( 4 );
    flags[ 0 ] = false;
    flags[ 1 ] = true;
    flags[ 2 ] = false;
    flags[ 3 ] = true;
    list< vector< FSP > > results;
    findall_proc( mach, extract_lexical_event, args, flags, results );
    for ( list< vector< FSP > >::iterator result_it = results.begin();
          result_it != results.end();
          ++result_it ) {
      string category;
      vector< string > event;
      if ( ! lilfes_to_c< string >().convert( mach, (*result_it)[ 0 ], category ) ) {
        throw UniMakerException( "2nd argument of extract_lexical_event/4 must be string" );
      }
      if ( ! lilfes_to_c< vector< string > >().convert( mach, (*result_it)[ 1 ], event ) ) {
        throw UniMakerException( "4th argument of extract_lexical_event/4 must be list of strings" );
      }
      string event_name;
      AmisModel::encode( category, event, event_name );
      events.push_back( event_name );
    }
  }
  if ( extract_lexical_event_feature_value ) {
    vector< FSP > args( 5 );
    args[ 0 ] = FSP( mach, model_name.c_str() );  // model name
    args[ 1 ] = FSP( mach );  // category
    args[ 2 ] = lexentry;     // lexical entry
    args[ 3 ] = FSP( mach );  // event (list of strings)
    args[ 4 ] = FSP( mach );  // feature value (float)
    vector< bool > flags( 5 );
    flags[ 0 ] = false;
    flags[ 1 ] = true;
    flags[ 2 ] = false;
    flags[ 3 ] = true;
    flags[ 4 ] = true;
    list< vector< FSP > > results;
    findall_proc( mach, extract_lexical_event_feature_value, args, flags, results );
    for ( list< vector< FSP > >::iterator result_it = results.begin();
          result_it != results.end();
          ++result_it ) {
      string category;
      vector< string > event;
      float feature_value;
      if ( ! lilfes_to_c< string >().convert( mach, (*result_it)[ 0 ], category ) ) {
        throw UniMakerException( "2nd argument of extract_lexical_event_feature_value/5 must be string" );
      }
      if ( ! lilfes_to_c< vector< string > >().convert( mach, (*result_it)[ 1 ], event ) ) {
        throw UniMakerException( "4th argument of extract_lexical_event_feature_value/5 must be list of strings" );
      }
      if ( ! lilfes_to_c< float >().convert( mach, (*result_it)[ 2 ], feature_value ) ) {
        throw UniMakerException( "5th argument of extract_lexical_event_feature_value/5 must be float" );
      }
      string event_name;
      AmisModel::encode( category, event, feature_value, event_name );
      events.push_back( event_name );
    }
  }
}

//////////////////////////////////////////////////////////////////////

// default version

void output_events( machine& mach, Lexbank& lexbank ) {
  // count lexical entries
  *error_stream << "Start counting lexical entries" << endl;
  FSArray< FSP , int > count_db( &mach );
  int num_sentences = 0;
  int num_counted_words = 0;
  int num_words = 0;
  for ( Lexbank::iterator sentence_it = lexbank.begin();
        sentence_it != lexbank.end();
        ++sentence_it ) {
    core_p trailp = mach.SetTrailPoint( NULL );
    if ( limit_sentences > 0 && sentence_it->first > limit_sentences ) break;
    ++num_sentences;
    for ( vector< FSP >::iterator lex_it = sentence_it->second.begin();
          lex_it != sentence_it->second.end();
          ++lex_it ) {
      Derivation term( *lex_it );
      if ( ! term.isTerminal() ) {
        throw UniMakerException( "lexbank contains inconsistent data" );
      }
      LexEntry lexentry( mach );
      if ( ( ! lexentry.lexWord().Unify( term.termWord() ) )
           || ( ! lexentry.lexTemplate().Unify( term.termTemplate() ) ) ) {
        throw UniMakerException( "making lexical entry failed.  maybe lilfes's bug??" );
      }
      int count = 0;
      if ( verbose_level > 1 ) { *error_stream << lex_it->DisplayAVM(); }
      if ( ! count_db.find( lexentry, count ) ) {
        count = 0;
        ++num_words;
      }
      if ( ! count_db.insert( lexentry, count + 1 ) ) {
        throw UniMakerException( "Cannot update lexicon count database" );
      }
      ++num_counted_words;
    }
    if ( verbose_level > 0 || num_sentences % 100 == 0 ) {
      *error_stream << "  # processed sentences = " << num_sentences << endl;
      *error_stream << "  # counted words = " << num_counted_words << endl;
      *error_stream << "  # words = " << num_words << endl;
    }
    mach.TrailBack( trailp );
  }
  *error_stream << "done." << endl;
  *error_stream << "  # processed sentences = " << num_sentences << endl;
  *error_stream << "  # counted words = " << num_counted_words << endl;
  *error_stream << "  # words = " << num_words << endl;

  // output unigram event data
  *error_stream << "Output event data" << endl;
  ogfstream uevent_file( uevent_file_name.c_str() );
  int event_id = 0;
  for ( FSArray< FSP, int >::iterator count_it = count_db.begin();
        count_it != count_db.end();
        ++count_it ) {
    //cerr << event_id << endl;
    core_p trailp = mach.SetTrailPoint( NULL );
    //cerr << "a1" << endl;
    //cerr << count_it.getArray() << " " << count_it.getLdbmIter().entry_id << endl;
    //cerr << count_db.end().getArray() << " " << count_db.end().getLdbmIter().entry_id << endl;
    LexEntry lexentry = count_it->first;
    //cerr << "a1.1" << endl;
    FSP word = lexentry.lexWord();
    FSP template_name = lexentry.lexTemplate();
    if ( ! word.IsValid() ) {
      throw UniMakerException( "Lexbank includes illegal data" );
    }
    vector< FSP > args( 2 );
    args[ 0 ] = word;
    args[ 1 ] = FSP( mach );
    if ( ! call_proc( mach, lookup_lexicon, args ) ) {
      throw UniMakerException( "lookup_lexicon failed" );
    }
    vector< FSP > template_list;
    if ( ! lilfes_to_c< vector< FSP > >().convert( mach, args[ 1 ], template_list ) ) {
      throw UniMakerException( "2nd argument of lookup_lexicon/2 is not a list" );
    }
    bool is_covered = false;
    for ( vector< FSP >::iterator temp_it = template_list.begin();
          temp_it != template_list.end();
          ++temp_it ) {
      if ( builtin::equivalent( mach, *temp_it, template_name ) ) {
        is_covered = true;
      }
    }
    if ( ! is_covered ) continue;
    uevent_file << "event_" << event_id << '\n';
    for ( vector< FSP >::iterator temp_it = template_list.begin();
          temp_it != template_list.end();
          ++temp_it ) {
      LexEntry tmp_lexentry( mach );
      tmp_lexentry.lexWord().Unify( word );
      tmp_lexentry.lexTemplate().Unify( *temp_it );
      if ( verbose_level > 1 ) { *error_stream << temp_it->DisplayAVM(); }
      vector< string > events;
      extract_events(mach, tmp_lexentry, events );
      if ( builtin::equivalent( mach, *temp_it, template_name ) ) {
        uevent_file << count_it->second << '\t';
      } else {
        uevent_file << "0\t";
      }
      for ( vector< string >::const_iterator event_it = events.begin();
            event_it != events.end();
            ++event_it ) {
        uevent_file << *event_it << ' ';
      }
      uevent_file << '\n';
    }
    uevent_file << '\n';
    ++event_id;
    if ( verbose_level > 0 || event_id % 1000 == 0 ) {
      *error_stream << "  # output events = " << event_id << endl;
    }
    mach.TrailBack( trailp );
  }
  *error_stream << "done." << endl;
}

//////////////////////////////////////////////////////////////////////

// feature forest version

void output_feature_forests( machine& mach, Lexbank& lexbank ) {
  *error_stream << "Start outputing feature forests" << endl;
  int num_events = 0;
  int num_words = 0;
  ogfstream uevent_file( uevent_file_name.c_str() );
  for ( Lexbank::iterator sentence_it = lexbank.begin();
        sentence_it != lexbank.end();
        ++sentence_it ) {
    core_p trailp = mach.SetTrailPoint( NULL );
    if ( limit_sentences > 0 && sentence_it->first > limit_sentences ) break;
    ++num_events;
    vector< FSP > word_list;
    vector< FSP > correct_list;
    vector< vector< FSP > > forest;
    bool is_covered = false;
    for ( vector< FSP >::iterator lex_it = sentence_it->second.begin();
          lex_it != sentence_it->second.end();
          ++lex_it ) {
      is_covered = false;
      if ( verbose_level > 1 ) { *error_stream << lex_it->DisplayAVM(); }
      Derivation term( *lex_it );
      if ( ! term.isTerminal() ) {
        throw UniMakerException( "lexbank contains inconsistent data" );
      }
      LexEntry lexentry( mach );
      FSP word = lexentry.lexWord();
      FSP template_name = lexentry.lexTemplate();
      if ( ( ! word.Unify( term.termWord() ) )
           || ( ! template_name.Unify( term.termTemplate() ) ) ) {
        throw UniMakerException( "making lexical entry failed.  maybe lilfes's bug??" );
      }
      // check coverage
      vector< FSP > args( 2 );
      args[ 0 ] = word;
      args[ 1 ] = FSP( mach );
      if ( ! call_proc( mach, lookup_lexicon, args ) ) {
        throw UniMakerException( "lookup_lexicon failed" );
      }
      vector< FSP > template_list;
      if ( ! lilfes_to_c< vector< FSP > >().convert( mach, args[ 1 ], template_list ) ) {
        throw UniMakerException( "2nd argument of lookup_lexicon/2 is not a list" );
      }
      for ( vector< FSP >::iterator temp_it = template_list.begin();
            temp_it != template_list.end();
            ++temp_it ) {
        if ( builtin::equivalent( mach, *temp_it, template_name ) ) {
          is_covered = true;
        }
      }
      if ( ! is_covered ) break;
      word_list.push_back( word );
      correct_list.push_back( template_name );
      forest.push_back( template_list );
    }

    if ( is_covered ) {
      uevent_file << "event_" << sentence_it->first << '\n';
      uevent_file << "1\t";
      vector< FSP >::iterator correct_it = correct_list.begin();
      for ( vector< FSP >::iterator word_it = word_list.begin();
            word_it != word_list.end();
            ++word_it, ++correct_it ) {
        LexEntry lexentry( mach );
        lexentry.lexWord().Unify( *word_it );
        lexentry.lexTemplate().Unify( *correct_it );
        vector< string > events;
        extract_events( mach, lexentry, events );
        for ( vector< string >::const_iterator event_it = events.begin();
              event_it != events.end();
              ++event_it ) {
          uevent_file << ' ' << *event_it;
        }
      }
      uevent_file << '\n';
      uevent_file << "{ _ ( _ ";
      vector< vector< FSP > >::iterator forest_it = forest.begin();
      for ( vector< FSP >::iterator word_it = word_list.begin();
            word_it != word_list.end();
            ++word_it, ++forest_it ) {
        ++num_words;
        uevent_file << "{ _ ";
        for ( vector< FSP >::iterator lex_it = forest_it->begin();
              lex_it != forest_it->end();
              ++lex_it ) {
          uevent_file << "( _ ";
          LexEntry lexentry( mach );
          lexentry.lexWord().Unify( *word_it );
          lexentry.lexTemplate().Unify( *lex_it );
          vector< string > events;
          extract_events( mach, lexentry, events );
          for ( vector< string >::const_iterator event_it = events.begin();
                event_it != events.end();
                ++event_it ) {
            uevent_file << *event_it << ' ';
          }
          uevent_file << ") ";
        }
        uevent_file << "} ";
      }
      uevent_file << ") }\n\n";
    }

    if ( verbose_level > 0 || num_events % 100 == 0 ) {
      *error_stream << "  # output events = " << num_events << endl;
      *error_stream << "  # words = " << num_words << endl;
    }
    mach.TrailBack( trailp );
  }
  *error_stream << "done." << endl;
}

//////////////////////////////////////////////////////////////////////

int main( int argc, char** argv ) {

  *error_stream << "unimaker (MAYZ " << MAYZ_VERSION << ")" << endl;
  *error_stream << MAYZ_COPYRIGHT << endl;

  lilfes::Initializer::PerformAll();

  // show help message
  if ( argc > 1 && std::strcmp( argv[ 1 ], "-h" ) == 0 ) {
    help_message();
    return 0;
  }

  // command-line arguments
  if ( ! analyze_arguments( argc, argv ) ) {
    help_message();
    return 1;
  }

  // Initialize machine object
  machine mach;

  try {

    // Load rule file
    *error_stream << "Loading feature file \"" << feature_file_name << "\"... ";
    if ( ! load_module( mach, "feature module", feature_file_name ) ) {
      *error_stream << "\nRule file \"" << feature_file_name << "\" not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    // Initialization
    *error_stream << "Initializing... ";
    if ( ! load_module( mach, "initialization module", "mayz/init" ) ) {
      *error_stream << "loading initialization module failed" << endl;
      return 1;
    }
    if ( ! mayz::Initializer::initialize( &mach ) ) {
      *error_stream << "initialization failed" << endl;
      return 1;
    }
    extract_lexical_event = prepare_proc( module::UserModule(), "extract_lexical_event", 4 );
    extract_lexical_event_feature_value = prepare_proc( module::UserModule(), "extract_lexical_event_feature_value", 5 );
    if ( ( ! extract_lexical_event ) && ( ! extract_lexical_event_feature_value ) ) {
      *error_stream << "neither 'extract_lexical_event/4' nor 'extract_lexical_event_feature_value/5' found" << endl;
      return 1;
    }
    if ( ! load_module( mach, "grammar module", "mayz/grammar" ) ) {
      *error_stream << "mayz/grammar module not found (maybe MAYZ Toolkit is not properly installed)" << endl;
      return 1;
    }
    import_lexicon = prepare_proc( module::UserModule(), "import_lexicon", 2 );
    if ( ! import_lexicon ) {
      *error_stream << "'import_lexicon/2' not found" << endl;
      return 1;
    }
    lookup_lexicon = prepare_proc( module::UserModule(), "lookup_lexicon", 2 );
    if ( ! lookup_lexicon ) {
      *error_stream << "'lookup_lexicon/2' not found" << endl;
      return 1;
    }
    *error_stream << "done." << endl;

  } catch ( exception& e ) {
    *error_stream << "Initialization failed: " << e.what() << endl;
    return 1;
  }

  *error_stream << "Output format: ";
  if ( feature_forest_mode ) {
    *error_stream << "AmisTree" << endl;
  } else {
    *error_stream << "Amis" << endl;
  }

  try {

    // Load data
    //   Lexicon lexicon( &mach );
    //   lexicon.load( lexicon_file_name );
    *error_stream << "Loading lexicon... ";
    vector< FSP > args( 2 );
    args[ 0 ] = FSP( mach, lexicon_file_name.c_str() );
    args[ 1 ] = FSP( mach, template_file_name.c_str() );
    if ( ! call_proc( mach, import_lexicon, args ) ) {
      *error_stream << "cannot import lexicon: " << lexicon_file_name << ", " << template_file_name << endl;
      return 1;
    }
    *error_stream << "done." << endl;
    *error_stream << "Loading lexbank... ";
    Lexbank lexbank( &mach );
    if ( ! lexbank.load( lexbank_file_name ) ) {
      *error_stream << "cannot load lexicon bank: " << lexbank_file_name << endl;
      return 1;
    }
    *error_stream << "done." << endl;

    if ( feature_forest_mode ) {
      output_feature_forests( mach, lexbank );
    } else {
      output_events( mach, lexbank );
    }

  } catch ( exception& e ) {
    cerr << "Fatal error: " << e.what() << endl;
    return 1;
  }

  Terminator::PerformAll();

  return 0;
}

